# Copyright (c) 2020-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import numpy as np
import pandas as pd

import h5py

from .exceptions import *
from . import vmap_structures


class VMAPImport:
    """The interface class to import a vmap file

    Parameters
    ----------
    filename : string
        The path to the vmap file to be read

    Raises
    ------
    Exception
        If the file cannot be read an exception is raised.
        So far any exception from the ``h5py`` module is passed through.
    """

    def __init__(self, filename):
        self._file = h5py.File(filename, 'r')
        self._mesh = None
        self._geometry = None
        self._state = None

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        pass

    def geometries(self):
        """Returns a list of geometry strings of geometries present in the vmap data
        """
        return self._file["/VMAP/GEOMETRY"].keys()

    def states(self):
        """Returns a list of state strings of states present in the vmap data
        """
        return self._file["/VMAP/VARIABLES/"].keys()

    def node_sets(self, geometry):
        """Returns a list of the node_sets present in the vmap file
        """
        return self._geometry_sets(geometry, 'nsets').keys()

    def element_sets(self, geometry):
        """Returns a list of the element_sets present in the vmap file
        """
        return self._geometry_sets(geometry, 'elsets').keys()

    def nodes(self, geometry):
        """Retrieves the node positions

        Parameters
        ----------
        geometry : string
            The geometry defined in the vmap file

        Returns
        -------
        node_positions : DataFrame
            a DataFrame with the node numbers as index and the columns 'x', 'y' and 'z' for the
            node coordinates.

        Raises
        ------
        KeyError
            if the geometry is not found of if the vmap file is corrupted
        """
        return pd.DataFrame(
            self._file["/VMAP/GEOMETRY/%s/POINTS/MYCOORDINATES" % geometry][()],
            columns=['x', 'y', 'z'],
            index=self._node_index(geometry)
        )

    def make_mesh(self, geometry, state=None):
        """Makes the initial mesh

        Parameters
        ----------
        geometry : string
            The geometry defined in the vmap file
        state : string, optional
            The load state of which the field variable is to be read.
            If not given, the state must be defined in ``join_variable()``.

        Returns
        -------
        self

        Raises
        ------
        KeyError
            if the ``geometry`` is not found of if the vmap file is corrupted
        KeyError
            if the ``node_set`` or ``element_set`` is not found in the geometry.
        APIUseError
            if both, a ``node_set`` and an ``element_set`` are given

        Notes
        -----
        This methods defines the initial mesh to which coordinate data can be joined by ``join_coordinates()``
        and field variables can be joined by ``join_variable()``

        Examples
        --------
        Get the mesh data with the coordinates of geometry '1' and the stress tensor of 'STATE-2'

        >>> (pylife.vmap.VMAPImport('demos/plate_with_hole.vmap')
             .make_mesh('1', 'STATE-2')
             .join_coordinates()
             .join_variable('STRESS_CAUCHY')
             .to_frame()
                                    x         y    z        S11       S22  S33        S12  S13  S23
        element_id node_id
        1          1734     14.897208  5.269875  0.0  27.080811  6.927080  0.0 -13.687358  0.0  0.0
                   1582     14.555333  5.355806  0.0  28.319006  1.178649  0.0 -10.732705  0.0  0.0
                   1596     14.630658  4.908741  0.0  47.701195  5.512213  0.0 -17.866833  0.0  0.0
                   4923     14.726271  5.312840  0.0  27.699907  4.052865  0.0 -12.210032  0.0  0.0
                   4924     14.592996  5.132274  0.0  38.010101  3.345431  0.0 -14.299768  0.0  0.0
        ...                       ...       ...  ...        ...       ...  ...        ...  ...  ...
        4770       3812    -13.189782 -5.691876  0.0  36.527439  2.470588  0.0 -14.706686  0.0  0.0
                   12418   -13.560289 -5.278386  0.0  32.868889  3.320898  0.0 -14.260107  0.0  0.0
                   14446   -13.673285 -5.569107  0.0  34.291058  3.642457  0.0 -13.836027  0.0  0.0
                   14614   -13.389065 -5.709927  0.0  36.063541  2.828889  0.0 -13.774759  0.0  0.0
                   14534   -13.276068 -5.419206  0.0  33.804211  2.829817  0.0 -14.580153  0.0  0.0
        """
        self._mesh = pd.DataFrame(index=self._mesh_index(geometry))
        self._geometry = geometry
        self._state = state
        return self

    def filter_node_set(self, node_set):
        """Filters a node set out of the current mesh

        Parameters
        ----------
        node_set : string
            The node set defined in the vmap file as geometry set

        Returns
        -------
        self

        Raises
        ------
        APIUseError
            If the mesh has not been initialized using ``make_mesh()``
        """
        self._check_mesh_for_filtering()
        node_set_ids = self._node_set_ids(self._geometry, node_set)
        self._mesh = self._mesh[self._mesh.index.isin(node_set_ids, level='node_id')]
        return self

    def filter_element_set(self, element_set):
        """Filters a node set out of the current mesh

        Parameters
        ----------
        element_set : string, optional
            The element set defined in the vmap file as geometry set

        Returns
        -------
        self

        Raises
        ------
        APIUseError
            If the mesh has not been initialized using ``make_mesh()``
        """
        self._check_mesh_for_filtering()
        element_set_ids = self._element_set_ids(self._geometry, element_set)
        self._mesh = self._mesh[self._mesh.index.isin(element_set_ids, level='element_id')]
        return self

    def _check_mesh_for_filtering(self):
        if self._mesh is None:
            raise APIUseError("Need to make_mesh() before filtering node or element sets.")

    def join_coordinates(self):
        """Join the coordinates of the predefined geometry in the mesh

        Returns
        -------
        self

        Raises
        ------
        APIUseError
            If the mesh has not been initialized using ``make_mesh()``


        Examples
        --------
        Receive the mesh with the node coordinates

        >>> pylife.vmap.VMAPImport('demos/plate_with_hole.vmap').make_mesh('1').join_coordinates().to_frame()
                                    x         y    z
        element_id node_id
        1          1734     14.897208  5.269875  0.0
                   1582     14.555333  5.355806  0.0
                   1596     14.630658  4.908741  0.0
                   4923     14.726271  5.312840  0.0
                   4924     14.592996  5.132274  0.0
        ...                       ...       ...  ...
        4770       3812    -13.189782 -5.691876  0.0
                   12418   -13.560289 -5.278386  0.0
                   14446   -13.673285 -5.569107  0.0
                   14614   -13.389065 -5.709927  0.0
                   14534   -13.276068 -5.419206  0.0

        [37884 rows x 3 columns]

        """
        if self._mesh is None:
            raise APIUseError("Need to make_mesh() before joining the coordinates.")
        self._mesh = self._mesh.join(self.nodes(self._geometry))
        return self

    def to_frame(self):
        """Returns the mesh and resets the mesh

        Returns
        -------
        mesh : DataFrame
            The mesh data joined so far

        Raises
        ------
        APIUseError
            if there is no mesh present, i.e. make_mesh() has not been called yet
            or the mesh has been reset in the meantime.

        Notes
        -----
        This method resets the mesh, i.e. ``make_mesh()`` must be called again in order to
        fetch more mesh data in another mesh.
        """
        if self._mesh is None:
            raise(APIUseError("Need to make_mesh() before requesting a resulting frame."))
        ret = self._mesh
        self._mesh = None
        return ret

    def join_variable(self, var_name, state=None, column_names=None):
        """Joins a field output variable to the mesh

        Parameters
        ----------
        var_name : string
            The name of the field variables
        state : string, opional
            The load state of which the field variable is to be read
            If not given, the last defined state, either defined in ``make_mesh()``
            or defeined in ``join_variable()`` is used.
        column_names : list of string, optional
            The names of the columns names to be used in the DataFrame
            If not provided, it will be chosen according to the list shown below.
            The length of the list must match the dimension of the variable.

        Returns
        -------
        self

        Raises
        ------
        APIUseError
            if the mesh has not been initialized using ``make_mesh()``
        KeyError
            if the geometry, state or varname is not found of if the vmap file is corrupted
        KeyError
            if there are no column names given and known for the variable.
        ValueError
            if the length of the column_names does not match the dimension of the variable

        Notes
        -----
        The mesh must be initialized with ``make_mesh()``. The final DataFrame can be retrieved with ``to_frame()``.

        If the ``column_names`` argument is not provided the following column names are chosen

        * 'DISPLACEMENT': ``['dx', 'dy', 'dz']``
        * 'STRESS_CAUCHY': ``['S11', 'S22', 'S33', 'S12', 'S13', 'S23']``
        * 'E': ``['E11', 'E22', 'E33', 'E12', 'E13', 'E23']``

        If that fails a ``KeyError`` exception is risen.

        Examples
        --------
        Receiving the 'DISPLACEMENT' of 'STATE-1' , the stress and strain tensors of 'STATE-2'

        >>> (pylife.vmap.VMAPImport('demos/plate_with_hole.vmap')
             .make_mesh('1')
             .join_variable('DISPLACEMENT', 'STATE-1')
             .join_variable('STRESS_CAUCHY', 'STATE-2')
             .join_variable('E').to_frame())
                             dx   dy   dz        S11       S22  S33        S12  S13  S23       E11       E22  E33       E12  E13  E23
        element_id node_id
        1          1734     0.0  0.0  0.0  27.080811  6.927080  0.0 -13.687358  0.0  0.0  0.000119 -0.000006  0.0 -0.000169  0.0  0.0
                   1582     0.0  0.0  0.0  28.319006  1.178649  0.0 -10.732705  0.0  0.0  0.000133 -0.000035  0.0 -0.000133  0.0  0.0
                   1596     0.0  0.0  0.0  47.701195  5.512213  0.0 -17.866833  0.0  0.0  0.000219 -0.000042  0.0 -0.000221  0.0  0.0
                   4923     0.0  0.0  0.0  27.699907  4.052865  0.0 -12.210032  0.0  0.0  0.000126 -0.000020  0.0 -0.000151  0.0  0.0
                   4924     0.0  0.0  0.0  38.010101  3.345431  0.0 -14.299768  0.0  0.0  0.000176 -0.000038  0.0 -0.000177  0.0  0.0
        ...                 ...  ...  ...        ...       ...  ...        ...  ...  ...       ...       ...  ...       ...  ...  ...
        4770       3812     0.0  0.0  0.0  36.527439  2.470588  0.0 -14.706686  0.0  0.0  0.000170 -0.000040  0.0 -0.000182  0.0  0.0
                   12418    0.0  0.0  0.0  32.868889  3.320898  0.0 -14.260107  0.0  0.0  0.000152 -0.000031  0.0 -0.000177  0.0  0.0
                   14446    0.0  0.0  0.0  34.291058  3.642457  0.0 -13.836027  0.0  0.0  0.000158 -0.000032  0.0 -0.000171  0.0  0.0
                   14614    0.0  0.0  0.0  36.063541  2.828889  0.0 -13.774759  0.0  0.0  0.000168 -0.000038  0.0 -0.000171  0.0  0.0
                   14534    0.0  0.0  0.0  33.804211  2.829817  0.0 -14.580153  0.0  0.0  0.000157 -0.000035  0.0 -0.000181  0.0  0.0

        [37884 rows x 15 columns]

        TODO
        ----
        Write a more central document about pyLife's column names.
        """
        if self._mesh is None:
            raise APIUseError("Need to make_mesh() before joining a variable.")
        if state is None:
            state = self._state
        if state is None:
            raise APIUseError("No state name given.\n"
                              "Must be either given in make_mesh() or in join_variable() as optional state argument.")
        self._state = state
        variable_data = (pd.DataFrame(index=self._mesh.index)
                         .join(self._variable(self._geometry, state, var_name, column_names)))
        self._mesh = self._mesh.join(variable_data.loc[self._mesh.index])
        return self

    def _mesh_index(self, geometry):
        connectivity = self._element_connectivity(geometry).connectivity
        length = sum([el.shape[0] for el in connectivity])
        index_np = np.empty((2, length), dtype=np.int64)

        i = 0
        for element_id, node_ids in connectivity.iteritems():
            i_next = i + node_ids.shape[0]
            index_np[0, i:i_next] = element_id
            index_np[1, i:i_next] = node_ids
            i = i_next

        return pd.MultiIndex.from_arrays(index_np, names=['element_id', 'node_id'])

    def _variable(self, geometry, state, var_name, column_names):
        if column_names is None:
            try:
                column_names = vmap_structures.column_names[var_name][0]
            except KeyError:
                raise KeyError("No column name for variable %s. Please provide with column_names parameter." % var_name)

        var_tree = self._file["/VMAP/VARIABLES/%s/%s/%s" % (state, geometry, var_name)]
        var_dimension = var_tree.attrs['MYDIMENSION']
        if len(column_names) != var_dimension:
            raise ValueError("Length of column name list (%d) does not match variable dimension (%d)."
                             % (len(column_names), var_dimension))

        return pd.DataFrame(
            data=var_tree['MYVALUES'][()],
            columns=column_names,
            index=self._make_index(var_tree, geometry)
        )

    def _element_connectivity(self, geometry):
        elements = self._file['/VMAP/GEOMETRY/' + geometry + '/ELEMENTS/MYELEMENTS']
        element_connectivity = elements['myIdentifier', 'myConnectivity'][:, 0]
        element_ids = [elid for elid, _ in element_connectivity]
        connectivity = [conn for _, conn in element_connectivity]
        return pd.DataFrame(data={'element_id': element_ids, 'connectivity': connectivity}).set_index('element_id')

    def _node_index(self, geometry):
        return pd.Index(
            self._file["/VMAP/GEOMETRY/%s/POINTS/MYIDENTIFIERS" % geometry][:, 0],
            name='node_id'
        )

    def _make_index(self, var_tree, geometry):
        location = var_tree.attrs['MYLOCATION']
        if location == 2:
            return self._var_node_index(var_tree)
        if location == 3:
            return self._var_element_index(var_tree)
        if location == 6:
            return self._var_element_nodal_index(var_tree, geometry)
        raise FeatureNotSupportedError("Unsupported value location, sorry\nSupported: NODE, ELEMENT, ELEMENT NODAL")

    def _var_node_index(self, var_tree):
        return pd.Index(var_tree['MYGEOMETRYIDS'][:, 0], name='node_id')

    def _var_element_index(self, var_tree):
        return pd.Index(var_tree['MYGEOMETRYIDS'][:, 0], name='element_id')

    def _var_element_nodal_index(self, var_tree, geometry):
        mesh_index_frame = self._mesh_index(geometry).to_frame(index=False)
        index_frame = pd.DataFrame(var_tree['MYGEOMETRYIDS'], columns=['element_id'])

        return (index_frame
                .merge(mesh_index_frame)
                .set_index(['element_id', 'node_id'])
                .index)

    def _geometry_sets(self, geometry, set_type):
        s_type = 0 if set_type == 'nsets' else 1
        geometry_sets = self._file["/VMAP/GEOMETRY/%s/GEOMETRYSETS" % geometry]
        return {
            gset.attrs['MYSETNAME'].decode('UTF-8'): gset['MYGEOMETRYSETDATA'][()]
            for (_, gset) in geometry_sets.items() if gset.attrs['MYSETTYPE'] == s_type
        }

    def try_get_geometry_set(self, geometry_name, geometry_set_name):
        try:
            geometry_set = self._file["/VMAP/GEOMETRY/%s/GEOMETRYSETS/%s/MYGEOMETRYSETDATA"
                                      % (geometry_name, geometry_set_name)]
            return pd.Index(geometry_set[()].flatten())
        except KeyError:
            return None

    def try_get_vmap_object(self, group_full_path):
        try:
            return self._file[group_full_path]
        except KeyError:
            return None

    def _node_set_ids(self, geometry, node_set):
        try:
            return self._geometry_sets(geometry, 'nsets')[node_set].T[0]
        except KeyError:
            raise KeyError("Node set '%s' not found in geometry '%s'" % (node_set, geometry))

    def _element_set_ids(self, geometry, element_set):
        try:
            return self._geometry_sets(geometry, 'elsets')[element_set].T[0]
        except KeyError:
            raise KeyError("Element set '%s' not found in geometry '%s'" % (element_set, geometry))



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Mustapha Kassem"
__maintainer__ = "Johannes Mueller"

import matplotlib
import matplotlib.pyplot as plt
import numpy as np

from scipy import stats


class ProbabilityDataDiagram:
    def __init__(self, probability_data, **kw):
        self._prob_data = probability_data
        self._occurrences_name = 'Occurrences' if 'occurrences_name' not in kw.keys() else kw['occurrences_name']
        self._probability_name = 'Probability' if 'probability_name' not in kw.keys() else kw['probability_name']
        self._title = 'Probability plot' if 'title' not in kw.keys() else kw['title']

    @property
    def occurrences_name(self):
        return self._occurrences_name

    @occurrences_name.setter
    def occurrences_name(self, name):
        self._occurrences_name = name

    @property
    def probability_name(self):
        return self._probability_name

    @probability_name.setter
    def probability_name(self, name):
        self._probability_name = name

    @property
    def title(self):
        return self._title

    @title.setter
    def title(self, name):
        self._title = name

    def __default_ax_config(self):
        fig, ax = plt.subplots()

        yticks = [1, 5, 10, 20, 50, 80, 90, 95, 99]
        ax.set_yticks([stats.norm.ppf(i/100.) for i in yticks])
        ax.set_yticklabels([str(i)+' %' for i in yticks])

        ax.set_xscale('log')
        ax.grid(True)

        ax.set_xlabel(self._occurrences_name)
        ax.set_ylabel(self._probability_name)

        ax.set_title(self._title)
        return ax

    def plot(self, ax=None):
        ax_local = self.__default_ax_config() if ax is None else ax
        ax_local.plot(self._prob_data.occurrences, self._prob_data.percentiles, 'ro')
        ax_local.plot([10**((i-self._prob_data.intercept)/self._prob_data.slope) for i in np.arange(-2.5, 2.5, 0.1)], np.arange(-2.5, 2.5, 0.1), 'r')

        return ax_local



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Mustapha Kassem"
__maintainer__ = "Johannes Mueller"

import numpy as np
import pandas as pd

from .meshsignal import MeshAccessor


@pd.api.extensions.register_dataframe_accessor('gradient')
class Gradient(MeshAccessor):
    '''Computes the gradient of a value in a 3D mesh

    Accesses a `mesh` registered in :mod:`meshsignal`

    Raises
    ------
    AttributeError
        if at least one of the columns `x`, `y` is missing
    AttributeError
        if the index of the DataFrame is not a two level MultiIndex
        with the names `node_id` and `element_id`


    Notes
    -----

    The gradient is calculated by fitting a plane into the nodes of
    each coordinate and the neigbor nodes usine least square fitting.

    The method is described in a `thread on stackoverflow`_.

    .. _thread on stackoverflow:  https://math.stackexchange.com/questions/2627946/how-to-approximate-numerically-the-gradient-of-the-function-on-a-triangular-mesh#answer-2632616
    '''
    def _find_neighbor(self):
        self.neighbors = {}

        sl = pd.IndexSlice
        for node, df in self._obj.groupby('node_id'):
            elidx = df.index.get_level_values('element_id')
            nodes = self._obj.loc[sl[:, elidx], :].index.get_level_values('node_id').to_numpy()
            self.neighbors[node] = np.setdiff1d(np.unique(nodes), [node])

    def _calc_lst_sqr(self):
        self.lst_sqr_grad_dx = {}
        self.lst_sqr_grad_dy = {}
        self.lst_sqr_grad_dz = {}
        groups = self._obj.groupby('node_id')
        self._node_data = np.zeros((len(groups), 4), order='F')
        self._node_data[:, :3] = groups.first()[['x', 'y', 'z']]
        self._node_data[:, 3] = groups[self.value_key].mean()
        for node, row in zip(groups.first().index, np.nditer(self._node_data.T, flags=['external_loop'], order='F')):
            diff = self._node_data[self.neighbors[node]-1, :] - row
            dx, dy, dz = np.linalg.lstsq(diff[:, :3], diff[:, 3], rcond=None)[0]
            self.lst_sqr_grad_dx[node] = dx
            self.lst_sqr_grad_dy[node] = dy
            self.lst_sqr_grad_dz[node] = dz

    def gradient_of(self, value_key):
        ''' returns the gradient

        Parameters
        ----------
        value_key : str
            The key of the value that forms the gradient. Needs to be found in ``df``

        Returns
        -------
        gradient : pd.DataFrame
            A table describing the gradient indexed by ``node_id``.
            The keys for the components of the gradients are
            ``['dx', 'dy', 'dz']``.
        '''
        self.value_key = value_key

        self.nodes_id = np.unique(self._obj.index.get_level_values('node_id'))

        self._find_neighbor()
        self._calc_lst_sqr()

        df_grad = pd.DataFrame({'node_id': self.nodes_id,
                                'dx': [*self.lst_sqr_grad_dx.values()],
                                'dy': [*self.lst_sqr_grad_dy.values()],
                                'dz': [*self.lst_sqr_grad_dz.values()]})
        df_grad = df_grad.set_index('node_id')

        return df_grad



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy as np
import pandas as pd
from scipy import optimize as op


class psdSignal:
    '''Handles different routines for self signals

    Remark: We are using the pandas data frame schema. The index contains the
    discrete frequency step. Every single column one self.

    Some functions of these class:

    * psd_optimizer
    * ...

    '''
    def __init__(self,df):

        self.df = df

    def rms_psd(self):
        f  = np.logspace(np.log10(self.index.values.min()),np.log10(
                              self.index.values.max()),2048)
        # f  = np.linspace(self.index.values.min(),self.index.values.max(),2048)
        psd = pd.DataFrame()
        for colact in self.columns:
            psd[colact] = np.interp(f,self.index.values,self[colact])
        psd.index = f
        return ((psd.diff()+psd).dropna()).multiply(np.diff(psd.index.values),axis = 0).sum()**0.5

    def _intMinlog(self,psdin,fsel,factor_rms_nods):
        self_rms_df = pd.DataFrame(data = 10**np.interp(psdin.index.values, fsel,np.log10(self)),
                                                    index = psdin.index.values)
        Ysel =  np.interp(fsel,psdin.index.values,psdin.values.flatten())
        rms_in = psdSignal.rms_psd(psdin).values
        rms_smooth = psdSignal.rms_psd(self_rms_df).values
        eps1 = (rms_in-rms_smooth)**2/rms_in**2
        eps2 =  np.dot(np.log10(Ysel/self),np.log10(Ysel/self))/np.dot(np.log10(Ysel),np.log10(Ysel))
        return factor_rms_nods*eps1+(1-factor_rms_nods)*eps2


    def psd_smoother(self,fsel,factor_rms_nodes = 0.5):
        ''' Smoothen a PSD using nodes and a penalty factor weighting the errors
        for the RMS and for the node PSD values


        Parameters
        ----------

        self: DataFrame
            unsmoothed PSD
        fsel: list or np.array
           nodes
        factor_rms_nodes: float (0 <= factor_rms_nods <= 1)
            penalty error weighting the errors:

            * 0: only error of node PSD values is considered
            * 1: only error of the RMS is considered


        Returns
        -------
        DataFrame
        '''

        f  = np.logspace(np.log10(self.index.values.min()),np.log10(
                              self.index.values.max()),1024)
        fsel = np.unique(fsel)
        fout = np.append(np.append(self.index.values.min(),np.unique(fsel)),self.index.values.max())
        opt_df = pd.DataFrame()
        for colact in self.columns:
            df_in = pd.DataFrame(data = np.interp(f,self.index.values,self[colact]),
                                 index = f)
            Hi0 = 10**(np.interp(fsel,f,np.log10(df_in.values.flatten())))
            # Hi = op.fmin(psdSignal._intMinlog,Hi0,args=(df_in,fsel,factor_rms_nodes),disp = 0)
            lim = np.array([df_in.values.min()*np.ones_like(fsel),
                            np.array(df_in.values.max()*np.ones_like(fsel))]).T

            Hi = op.minimize(psdSignal._intMinlog,x0 = Hi0,bounds = tuple(map(tuple, lim)),
                             args=(df_in,fsel,factor_rms_nodes))
            opt_df[colact] =  10**np.interp(fout,fsel,np.log10(Hi.x))

        # opt_df.index = f
        opt_df.index = fout
        return opt_df



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Gyöngyvér Kiss"
__maintainer__ = __author__

import ipywidgets as widgets
from IPython.display import display

from ..analyzers.maxlike import MaxLikeInf, MaxLikeFull
from ..analyzers.probit import Probit
from ..analyzers.bayesian import Bayesian
from .radio_button_woehler_curve import RadioButtonWoehlerCurve


class WoehlerCurveAnalyzerOptions(RadioButtonWoehlerCurve):
    def __init__(self, df):
        self.woehler_curve = None
        self._df = df
        self.collect_fixed_params()
        self._option_dict = {
            'Maximum likelihood 2 params': (MaxLikeInf(self._df), self.__just_analyze),
            'Maximum likelihood 5 params': (MaxLikeFull(self._df), self.__display_param_fix),
            'Probit': (Probit(self._df), self.__just_analyze),
            'Bayes (SLOW!)': (Bayesian(self._df), self.__just_analyze)
        }

        super().__init__(self._option_dict.keys(), 'Select method')
        self.calculate_curve_button = widgets.Button(description='Calculate curve')
        self.calculate_curve_button.on_click(self.calculate_curve_button_clicked_handler)
        self._analyzer, _ = list(self._option_dict.values())[0]
        self.__just_analyze()

    def analyzer(self):
        return self._analyzer

    def __just_analyze(self, **kwargs):
        self.woehler_curve = self._analyzer.analyze(**kwargs)
        print(self.woehler_curve)
        self.bic = self._analyzer.bayesian_information_criterion()
        print("BIC:", self.bic)

    def __display_param_fix(self):
        display(self.param_fix_tab)
        display(self.calculate_curve_button)

    def selection_changed_handler(self, change):
        self.clear_selection_change_output()
        try:
            self._analyzer, analyze_method = self._option_dict[change['new']]
        except KeyError:
            raise AttributeError('Unexpected selection')

        analyze_method()

    def tab_content_changed_handler(self, change):
        try:
            self.param_fix.update({change.owner.description: float(change.new)})
        except(ValueError, TypeError):
            self.param_fix.update({change.owner.description: ''})

    def collect_fixed_params(self):
        self.param_fix = {'SD_50': '', '1/TS': '', 'ND_50': '', 'k_1': '', '1/TN': ''}
        items = []
        for k in self.param_fix:
            text = widgets.Text(description=k, value=self.param_fix[k])
            text.observe(self.tab_content_changed_handler, names='value')
            items.append(text)
        self.param_fix_tab = widgets.VBox(items)

    def calculate_curve_button_clicked_handler(self, b):
        param_fix = {k: v for k, v in self.param_fix.items() if v != ''}
        self.__just_analyze(fixed_parameters=param_fix)



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Gyöngyvér Kiss"
__maintainer__ = __author__

from ..diagrams.woehler_curve_diagrams import WoehlerCurveDiagrams
from .radio_button_woehler_curve import RadioButtonWoehlerCurve


class WholeWoehlerCurvePlotter(RadioButtonWoehlerCurve):
    def __init__(self, woehler_curve, fatigue_data, analyzer):
        self.woehler_curve = woehler_curve
        self.woehler_curve_diagrams = WoehlerCurveDiagrams(self.woehler_curve, fatigue_data, analyzer)
        super().__init__(options=['k_2 = inf', 'k_2 = k_1', 'k_2 = 2 k_1 - 1'], description='Runout zone plot')
        self.woehler_curve_diagrams.plot_whole_woehler_curve_graph(0)

    def selection_changed_handler(self, change):
        self.clear_selection_change_output()
        if change['new'] == change.owner.options[0]:
            self.woehler_curve_diagrams.plot_whole_woehler_curve_graph(None)
        elif change['new'] == change.owner.options[1]:
            self.woehler_curve_diagrams.plot_whole_woehler_curve_graph(self.woehler_curve.k)
        elif change['new'] == change.owner.options[2]:
            self.woehler_curve_diagrams.plot_whole_woehler_curve_graph(2 * self.woehler_curve.k - 1)
        else:
            raise AttributeError('Unexpected selection')



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Cedric Philip Wagner"
__maintainer__ = "Johannes Mueller"

import numpy as np
import logging
import pandas as pd

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
ch = logging.StreamHandler()
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s\n')
ch.setFormatter(formatter)
ch.setLevel(logging.DEBUG)
logger.addHandler(ch)


class FiniteLifeBase:
    """Base class for SN curve calculations - either in logarithmic or regular scale"""

    def __init__(self, k_1, SD_50, ND_50):
        self._k = k_1
        self.SD_50 = SD_50
        self.ND_50 = ND_50
        self._check_inputs_valid(k_1=self.k_1,
                                 SD_50=self.SD_50,
                                 ND_50=self.ND_50)

    @property
    def k_1(self):
        return self._k

    @k_1.setter
    def k_1(self, k_1):
        self._k = k_1

    def _check_inputs_valid(self, **kwargs):
        try:
            for param, value in kwargs.items():
                if not float(value) > 0:
                    logger.error("Parameter '{}' is not a number greater than zero: {}".format(
                            param,
                            value,
                            ))
        except Exception as e:
            raise ValueError("Not all inputs are valid numbers: {}".format(e))


class FiniteLifeLine(FiniteLifeBase):
    """Sample points on the finite life line - either N or S (LOGARITHMIC SCALE)

    The formula for calculation is taken from
    "Betriebsfestigkeit", Haibach, 3. Auflage 2006

    Notes
    -----
    In contrast to the case

    Parameters
    ----------
    k : float
        slope of the SN-curve
    SD_50 : float
        lower stress limit in the finite life region
    ND_50 : float
        number of cycles at stress SD_50
    """

    def __init__(self, k, SD_50, ND_50):
        super(FiniteLifeLine, self).__init__(k, SD_50, ND_50)
        self.SD_50_log = np.log10(self.SD_50)
        self.ND_50_log = np.log10(self.ND_50)
        if self.k_1 < 0:
            logger.warning("Parameter k should be positive. Given k ('{}') has been "
                           "assigned as absolute value.".format(self.k_1))
            self.k_1 = abs(self.k_1)

    def calc_S_log(self, N_log, ignore_limits=False):
        """Calculate stress logarithmic S_log for a given number of cycles N_log

        Parameters
        ----------
        N_log : float
            logarithmic number of cycles
        ignore_limits : boolean
            ignores the upper limit of the number of cycles
            generally it should be smaller than ND_50 (=the limit of the finite life region)
            but some special evaluation methods (e.g. according to marquardt2004) require
            extrapolation to estimate an equivalent stress

        Returns
        -------
        S_log : float
            logarithmic stress corresponding to the given number of cycles (point on the SN-curve)
        """
        if N_log > self.ND_50_log:
            if ignore_limits:
                logger.warning("The limits to the interpolation of the finite life region "
                               "have explicitly been ignored.")
            else:
                raise ValueError("The given N_log ('{}') must be smaller than "
                                 "the lifetime ('{}').".format(
                                     N_log,
                                     self.ND_50_log,
                                     ))
        elif N_log < 1:
            raise ValueError("Invalid input for parameter N_log: {}".format(N_log))
        return self.SD_50_log + (self.ND_50_log - N_log) / self.k_1

    def calc_N_log(self, S_log, ignore_limits=False):
        """Calculate number of cycles N_log for a given stress S_log

        Parameters
        ----------
        S_log : float
            logarithmic stress (point on the SN-curve)
        ignore_limits : boolean
            ignores the upper limit of the number of cycles
            generally it should be smaller than ND_50_log (=the limit of the finite life region)
            but some special evaluation methods (e.g. according to marquardt2004) require
            extrapolation to estimate an equivalent stress

        Returns
        -------
        N_log : float
            logarithmic number of cycles corresponding to the given stress value (point on the SN-curve)
        """
        if S_log < self.SD_50_log:
            if ignore_limits:
                logger.warning("The limits to the interpolation of the finite life region "
                               "have explicitly been ignored.")
            else:
                raise ValueError("The given stress S_log ('{}') must be larger than "
                                 "SD_50_log ('{}').".format(
                                     S_log,
                                     self.SD_50_log,
                                     ))
        return self.ND_50_log + (self.SD_50_log - S_log) * self.k_1


class FiniteLifeCurve(FiniteLifeBase):
    """Sample points on the finite life curve - either N or S (NOT logarithmic scale)

    The formula for calculation is taken from
    "Betriebsfestigkeit", Haibach, 3. Auflage 2006

    **Consider:** load collective and life curve have to be consistent:

        * range vs range
        * amplitude vs amplitude

    Parameters
    ----------
    k : float
        slope of the SN-curve
    SD_50 : float
        lower stress limit in the finite life region
    ND_50 : float
        number of cycles at stress SD_50
    """
    def __init__(self, k_1, SD_50, ND_50):
        super(FiniteLifeCurve, self).__init__(k_1, SD_50, ND_50)

    def calc_S(self, N, ignore_limits=True):
        """Calculate stress S for a given number of cycles N

        Parameters
        ----------
        N : float
            number of cycles
        ignore_limits : boolean
            ignores the upper limit of the number of cycles
            generally it should be smaller than ND_50 (=the limit of the finite life region)
            but some special evaluation methods (e.g. according to marquardt2004) require
            extrapolation to estimate an equivalent stress

        Returns
        -------
        S : float
            stress corresponding to the given number of cycles (point on the SN-curve)
        """
        if N > self.ND_50:
            if ignore_limits:
                logger.warning("The limits to the interpolation of the finite life region "
                               "have explicitly been ignored.")
            else:
                raise ValueError("The given N ('{}') must be smaller than "
                                 "the lifetime ('{}').".format(
                                     N,
                                     self.ND_50,
                                     ))
        elif N < 1:
            raise ValueError("Invalid input for parameter N: {}".format(N))
        return self.SD_50 * (N / self.ND_50)**(- 1/self.k_1)

    def calc_N(self, S, ignore_limits=False):
        """Calculate number of cycles N for a given stress S

        Parameters
        ----------
        S : array like
            Stress (point(s) on the SN-curve)
        ignore_limits : boolean
            ignores the upper limit of the number of cycles
            generally it should be smaller than ND_50 (=the limit of the finite life region)
            but some special evaluation methods (e.g. according to marquardt2004) require
            extrapolation to estimate an equivalent stress

        Returns
        -------
        N : array like
            number of cycles corresponding to the given stress value (point on the SN-curve)
        """
        if np.any(S < self.SD_50):
            if ignore_limits:
                logger.warning("Using all cycles")
            else:
                raise ValueError("The given stress S ('{}') must be larger than "
                                 "SD_50 ('{}').".format(
                                     S,
                                     self.SD_50,
                                     ))
        return self.ND_50 * (S / self.SD_50)**(-self.k_1)

    def calc_damage(self, loads, method="elementar", index_name="range"):
        """Calculate the damage based on the methods
         * Miner elementar (k_2 = k)
         * Miner Haibach (k_2 = 2k-1)
         * Miner original (k_2 = -\inf)

        **Consider:** load collective and life curve have to be consistent:

        * range vs range
        * amplitude vs amplitude

        Parameters
        ----------
        loads : pandas series histogram
            loads (index is the load, column the cycles)
        method : str
         * 'elementar': Miner elementar (k_2 = k)
         * 'MinerHaibach': Miner Haibach (k_2 = 2k-1)
         * 'original': Miner original (k_2 = -\inf)

        Returns
        -------
        damage : pd.DataFrame
            damage for every load horizont based on the load collective and the method
        """
        damage = pd.DataFrame(index=loads.index,columns=loads.columns, data=0)
        load_values = loads.index.get_level_values(index_name).mid.values
        # Miner elementar
        cycles_SN = self.calc_N(load_values, ignore_limits=True)
        if method == 'original':
            damage = loads.divide(cycles_SN,axis = 0)
            damage[load_values <= self.SD_50] = 0
        elif method == 'elementar':
            damage = loads.divide(cycles_SN,axis = 0)
        elif method == 'MinerHaibach':
            # k2
            cycles_SN[load_values <= self.SD_50] = self.ND_50 * (load_values[load_values <= self.SD_50] / self.SD_50)**(-(2*self.k_1-1))
            damage = loads.divide(cycles_SN, axis=0)
        return damage



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Mustapha Kassem"
__maintainer__ = "Johannes Mueller"


import numpy as np
import matplotlib
import matplotlib.pyplot as plt

from .woehler_curve_graph import WoehlerCurveGraph
from .whole_woehler_curve_graph import WholeWoehlerCurveGraph


class WoehlerCurveDiagrams:
    def __init__(self, woehler_curve, fatigue_data = None, analyzer = None,
                 y_min=None, y_max=None, x_min=None, x_max=None, ax=None):
        self.woehler_curve = woehler_curve
        self._analyzer = analyzer
        self._fd = fatigue_data
        self.y_min = self._fd.load.min()*0.8 if y_min is None else y_min
        self.y_max = self._fd.load.max()*1.2 if y_max is None else y_max
        self.x_min = self._fd.cycles.min()*0.4 if x_min is None else x_min
        self.x_max = self._fd.cycles.max()*2 if x_max is None else x_max
        
        self.xlim_WL = (self.x_min,round(self.x_max, -1))
        self.ylim_WL = (self.y_min,round(self.y_max, -1))
        
        self.reset(ax)

    def reset(self, ax=None):
        if ax is None:
            _, self._ax = plt.subplots()
        else:
            self._ax = ax
        self.__default_ax_config()
        return self

    def __default_ax_config(self):
        self._ax.set_xlim(self.xlim_WL)
        self._ax.set_ylim(self.ylim_WL)
        self._ax.set_xscale('log')
        self._ax.set_yscale('log')
        self._ax.grid(True)
        self._ax.set_xlabel('Number of cycles', fontsize=11)
        self._ax.set_ylabel('Amplitude' + ' (' + 'Stress' + ') in ' + '$N/mm^2$' + '(log scaled)')
        matplotlib.rcParams.update({'font.size': 11})

    def plot_fatigue_data(self):
        self._ax.plot(self._fd.runouts.cycles,
                      self._fd.runouts.load, 'bo', mfc='none', label='Runout')
        self._ax.plot(self._fd.fractures.cycles,
                      self._fd.fractures.load, 'bo', label='Failure')
        # self.plot_endurance_limit()
        self._ax.legend(loc='upper right', fontsize=11)
        return self

    def plot_endurance_limit(self):
        self._ax.axhline(y=self._fd.fatigue_limit, linewidth=2, color='r', label='Endurance limit')
        return self

    def plot_woehler_curve(self, failure_probablility=0.5, **kw):
        woehler_curve_graph = WoehlerCurveGraph(self.woehler_curve, self.y_min, self.y_max, failure_probablility)

        self._ax.plot(woehler_curve_graph.points[:, 1],
                      woehler_curve_graph.points[:, 0], **kw)
        self._ax.legend(loc='upper right', fontsize=11)
        return self

    def plot_pearl_chain_method(self):
        pce = self._analyzer.pearl_chain_estimator()
        normed_load = pce.normed_load
        normed_cycles = pce.normed_cycles
        self._ax.plot(normed_cycles, np.ones(len(normed_cycles))*normed_load,
                      'go', label='PCM shifted probes', marker="v")
        self._ax.plot(self.xlim_WL, np.ones(len(self.xlim_WL))*normed_load, 'g')
        self.plot_woehler_curve()
        self._ax.legend(loc='upper right', fontsize=11)
        return self

    def plot_deviation(self, **kw):
        self.plot_woehler_curve(failure_probablility=0.5, color='r', label='$P_A$=50%')
        self.plot_woehler_curve(failure_probablility=0.1, color='r', linestyle='--', label='$P_A$=10% u. 90%')
        self.plot_woehler_curve(failure_probablility=0.9, color='r', linestyle='--')

        text = '$k$ = '+str(np.round(self.woehler_curve['k_1'], decimals=2)) + '\n'
        text += '$1/T_N$ = ' + str(np.round(self.woehler_curve['1/TN'], decimals=2)) + '\n'
        text += '$1/T_S^*$ = ' + str(np.round(self.woehler_curve['1/TS'], decimals=2))

        self._ax.text(0.01, 0.03, text, verticalalignment='bottom',
                    horizontalalignment='left', transform=self._ax.transAxes,
                    bbox={'facecolor': 'grey', 'alpha': 0.2, 'pad': 10})
        self._ax.legend(loc='upper right', fontsize=11)
        return self

    def plot_fitted_curve(self, k_2=None):
        ''' This is broken now
        '''
        whole_woehler_curve_graph = WholeWoehlerCurveGraph(self.woehler_curve, k_2, self.y_min, self.y_max)
        WL_50 = whole_woehler_curve_graph.graph_50
        WL_10 = whole_woehler_curve_graph.graph_10
        WL_90 = whole_woehler_curve_graph.graph_90
        self._ax.plot(WL_50[:, 1], WL_50[:, 0], 'r', linewidth=2., label=u'WC, $P_A$=50%')
        self._ax.plot(WL_10[:, 1], WL_10[:, 0], 'r', linewidth=1.5, linestyle='--', label=u'WC, $P_A$=10% u. 90%')
        self._ax.plot(WL_90[:, 1], WL_90[:, 0], 'r', linewidth=1.5, linestyle='--')
        self._ax.legend(loc='upper right', fontsize=11)

        text = '$k_1$ = '+str(np.round(self.woehler_curve['k_1'],decimals=2)) + '\n'
        text += '$1/T_N$ = ' + str(np.round(self.woehler_curve['1/TN'],decimals=2)) + '\n'
        text += '$1/T_S^*$ = ' + str(np.round(self.woehler_curve['1/TN']**(1./self.woehler_curve['k_1']), decimals=2)) + '\n'
        text += '$S_{D,50}$ = ' + str(np.round(self.woehler_curve['SD_50'],decimals=1)) + '\n'
        text += '$N_{D,50}$ = ' + '{:1.2e}'.format(self.woehler_curve['ND_50']) + '\n'
        text += '$1/T_S$ = ' + str(np.round(self.woehler_curve['1/TS'],decimals=2))

        self._ax.text(0.01, 0.03, text,
                 verticalalignment='bottom',horizontalalignment='left',
                 transform=self._ax.transAxes, bbox={'facecolor':'grey', 'alpha':0.2, 'pad':10})
        return self

    def plot_slope(self, failure_probablility=0.5):
        text = '$k$ = '+str(np.round(self.woehler_curve['k_1'], decimals=2))
        self._ax.text(0.01, 0.03, text, verticalalignment='bottom',
            horizontalalignment='left', transform=self._ax.transAxes,
            bbox={'facecolor': 'grey', 'alpha': 0.2, 'pad': 10})
        self.plot_fatigue_data()
        self.plot_woehler_curve()
        return self



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Mustapha Kassem"
__maintainer__ = "Johannes Mueller"

import numpy as np
import pandas as pd
from scipy import stats

from .woehler_curve_graph import WoehlerCurveGraph

class WholeWoehlerCurveGraph:

    def __init__(self, woehler_curve, k_2, y_min, y_max):
        self.woehler_curve = woehler_curve
        self.k_2 = woehler_curve['k_1'] if k_2 is None else k_2
        self.y_min = y_min
        self.y_max = y_max

        SD_50 = woehler_curve['SD_50']
        SD_10 = SD_50 / (10**(-stats.norm.ppf(0.1)*np.log10(woehler_curve['1/TS'])/2.56))
        SD_90 = SD_50 / (10**(-stats.norm.ppf(0.9)*np.log10(woehler_curve['1/TS'])/2.56))

        self.graph_50 = self.__create_whole_woehler_curve_graph(SD_50, 0.5)
        self.graph_10 = self.__create_whole_woehler_curve_graph(SD_10, 0.1)
        self.graph_90 = self.__create_whole_woehler_curve_graph(SD_90, 0.9)


    def __create_whole_woehler_curve_graph(self, y_lim, pa_goal):
        woehler_curve = self.woehler_curve
        graph_1 = WoehlerCurveGraph(woehler_curve, y_lim, self.y_max, pa_goal)

        woehler_curve_2 = pd.Series({
            'SD_50': y_lim, '1/TS': woehler_curve['1/TS'], 'ND_50': graph_1.points[-1, -1],
            'k_1': self.k_2, '1/TN': woehler_curve['1/TN']})
        graph_2 = WoehlerCurveGraph(woehler_curve_2, self.y_min, y_lim)
        return np.append(graph_1.points, graph_2.points, axis=0)



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Mustapha Kassem"
__maintainer__ = "Johannes Mueller"

import numpy as np
import pandas as pd
import theano.tensor as tt
import pymc3 as pm

from .elementary import Elementary
from .likelihood import Likelihood


class Bayesian(Elementary):

    class _LogLike(tt.Op):
        """
        Specify what type of object will be passed and returned to the Op when it is
        called. In our case we will be passing it a vector of values (the parameters
        that define our model) and returning a single "scalar" value (the
        log-likelihood)

        http://mattpitkin.github.io/samplers-demo/pages/pymc3-blackbox-likelihood/
        """
        itypes = [tt.dvector]  # expects a vector of parameter values when called
        otypes = [tt.dscalar]  # outputs a single scalar value (the log likelihood)

        def __init__(self, likelihood):
            """
            Initialise the Op with various things that our log-likelihood function
            requires. Below are the things that are needed in this particular
            example.

            Parameters
            ----------
            loglike:
                The log-likelihood function we've defined
            data:
                The "observed" data that our log-likelihood function takes in
            x:
                The dependent variable (aka 'load_cycle_limit') that our model requires

            """
            # add inputs as class attributes
            self.likelihood = likelihood

        def perform(self, node, inputs, outputs):
            # the method that is used when calling the Op
            var, = inputs  # this will contain my variables
            # mali_sum_lolli(var, self.zone_inf, self.load_cycle_limit):
            # call the log-likelihood function
            logl = self.likelihood.likelihood_infinite(var[0], var[1])

            outputs[0][0] = np.array(logl)  # output the log-likelihood

    def _specific_analysis(self, wc, nsamples=1000, **kw):
        self._nsamples = nsamples
        nburn = self._nsamples // 10

        tune = kw.pop('tune', 1000)
        random_seed = kw.pop('random_seed', None)

        slope_trace = self._slope_trace(tune=tune, random_seed=random_seed, **kw)
        TN_trace = self._TN_trace(tune=tune, random_seed=random_seed, **kw)
        SD_TS_trace = self._SD_TS_trace(tune=tune, random_seed=random_seed, **kw)

        slope = slope_trace.get_values('x')[nburn:].mean()
        intercept = slope_trace.get_values('Intercept')[nburn:].mean()
        SD_50 = SD_TS_trace.get_values('SD_50')[nburn:].mean()
        ND_50 = np.power(10., np.log10(SD_50) * slope + intercept)

        res = {
            'SD_50': SD_50,
            '1/TS': SD_TS_trace.get_values('TS_50')[nburn:].mean(),
            'ND_50': ND_50,
            'k_1': -slope,
            '1/TN': TN_trace.get_values('mu')[nburn:].mean(),
        }

        return pd.Series(res)

    def _slope_trace(self, chains=2, random_seed=None, tune=1000, **kw):
        data_dict = {
            'x': np.log10(self._fd.fractures.load),
            'y': np.log10(self._fd.fractures.cycles.to_numpy())
        }
        with pm.Model():
            family = pm.glm.families.StudentT()
            pm.glm.GLM.from_formula('y ~ x', data_dict, family=family)
            trace_robust = pm.sample(self._nsamples,
                                     target_accept=0.99,
                                     random_seed=random_seed,
                                     chains=chains,
                                     tune=tune,
                                     **kw)

            return trace_robust

    def _TN_trace(self, chains=3, random_seed=None, tune=1000, **kw):
        with pm.Model():
            log_N_shift = np.log10(self._pearl_chain_estimator.normed_cycles)
            stdev = pm.HalfNormal('stdev', sigma=1.3)  # sigma standard wert (log-normal/ beat Verteilung/exp lambda)
            mu = pm.Normal('mu', mu=log_N_shift.mean(), sigma=log_N_shift.std())
            _ = pm.Normal('y', mu=mu, sigma=stdev, observed=log_N_shift)

            trace_TN = pm.sample(self._nsamples,
                                 target_accept=0.99,
                                 random_seed=random_seed,
                                 chains=chains,
                                 tune=tune,
                                 **kw)

        return trace_TN

    def _SD_TS_trace(self, chains=3, random_seed=None, tune=1000, **kw):
        loglike = self._LogLike(Likelihood(self._fd))

        with pm.Model():
            inf_load = self._fd.infinite_zone.load
            SD = pm.Normal('SD_50', mu=inf_load.mean(), sigma=inf_load.std()*5)
            TS = pm.Lognormal('TS_50', mu=np.log10(1.1), sigma=np.log10(0.5))

            # convert m and c to a tensor vector
            var = tt.as_tensor_variable([SD, TS])

            pm.Potential('likelihood', loglike(var))

            trace_SD_TS = pm.sample(self._nsamples,
                                    cores=1,
                                    chains=chains,
                                    random_seed=random_seed,
                                    discard_tuned_samples=True,
                                    tune=tune,
                                    **kw)

        return trace_SD_TS



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Rainflow counting
=================

A module performing rainflow counting
'''

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import numpy as np
import pandas as pd

import cython


def get_turns(samples):
    ''' Finds the turning points in a sample chunk

    Parameters
    ----------
    samples : 1D numpy.ndarray
        the sample chunk

    Returns
    -------
    positions : 1D numpy.ndarray
        the indeces where sample has a turning point
    turns : 1D numpy.ndarray
        the values of the turning points

    '''
    def plateau_turns(diffs):
        plateau_turns = np.zeros_like(diffs, dtype=np.bool_)[1:]
        duplicates = np.array(diffs == 0, dtype=np.int8)

        if duplicates.any():
            edges = np.diff(duplicates)
            dups_starts = np.where(edges > 0)[0]
            dups_ends = np.where(edges < 0)[0]
            if len(dups_starts) and len(dups_ends) and dups_ends[0] < dups_starts[0]:
                dups_ends = dups_ends[1:]
            plateau_turns[dups_starts[np.where(diffs[dups_starts] * diffs[dups_ends+1] < 0)]] = True

        return plateau_turns

    diffs = np.diff(samples)
    peak_turns = diffs[:-1] * diffs[1:] < 0.0

    positions = np.where(np.logical_or(peak_turns, plateau_turns(diffs)))[0] + 1

    return positions, samples[positions]


class AbstractRainflowCounter:
    '''The common base class for rainflow counters

    Subclasses implementing a specific rainflow counting algorithm are
    supposed to implement a method ``process()`` that takes the signal
    samples as a parameter, append all the hysteresis loop limits to
    ``self._loops_from`` and ``self.loops_to`` and return ``self``. The
    ``process()`` method is supposed to be implemented in a way, that
    the result is independent of the sample chunksize, so
    ``rfc.process(signal)`` should be equivalent to
    ``rfc.process(signal[:n]).process(signal[n:])`` for any 0 < n <
    signal length.

    Todo
    ----
    - write a 4 point rainflow counter
    - accept the histogram binning upfront so that loop information
      has not to be stored explicitly. This is important to ensure
      that the memory consumption remains O(1) rather than O(n).
      '''
    def __init__(self):
        self.loops_from = []
        self.loops_to = []
        self._sample_tail = None

    def _get_new_turns(self, samples):
        if self._sample_tail is not None:
            samples = np.concatenate((self._sample_tail, samples))
        turn_positions, turns = get_turns(samples)
        if turn_positions.size > 0:
            self._sample_tail = samples[turn_positions[-1]:]
        else:
            self._sample_tail = samples
        return turns

    def get_rainflow_matrix(self, bins):
        ''' Calculates a histogram of the recorded loops

        Parameters
        ----------
        bins : int or array_like or [int, int] or [array, array], optional
            The bin specification (see numpy.histogram2d)

        Returns
        -------
        H : ndarray, shape(nx, ny)
            The bi-dimensional histogram of samples (see numpy.histogram2d)
        xedges : ndarray, shape(nx+1,)
            The bin edges along the first dimension.
        yedges : ndarray, shape(ny+1,)
            The bin edges along the second dimension.
        '''
        return np.histogram2d(self.loops_from, self.loops_to, bins)

    def get_rainflow_matrix_frame(self, bins):
        '''Calculates a histogram of the recorded loops into a pandas.DataFrame.

        An interval index is used to index the bins.

        Parameters
        ----------
        bins : int or array_like or [int, int] or [array, array], optional
            The bin specification: see numpy.histogram2d

        Returns
        -------
        pandas.DataFrame
            A pandas.DataFrame using a multi interval index in order to
            index data point for a given from/to value pair.
        '''
        hist, fr, to = self.get_rainflow_matrix(bins)
        index_fr = pd.IntervalIndex.from_breaks(fr)
        index_to = pd.IntervalIndex.from_breaks(to)

        mult_idx = pd.MultiIndex.from_product([index_fr, index_to], names=['from', 'to'])
        return pd.DataFrame(data=hist.flatten(), index=mult_idx)

    def residuals(self):
        ''' Returns the residual turning points of the time signal so far

        The residuals are the loops not (yet) closed.
        '''
        return self._residuals


class RainflowCounterThreePoint(AbstractRainflowCounter):
    ''' Implements 3 point rainflow counting algorithm

    See the `here <subsection_TP_>`_ in the demo for an example.

    We take three turning points into account to detect closed hysteresis loops.

    * start: the point where the loop is starting from
    * front: the turning point after the start
    * back: the turning point after the front

    A loop is considered closed if following conditions are met:

    * the load difference between front and back is bigger than or
      equal the one between start and front. In other words: if the
      back goes beyond the starting point. For example (A-B-C) and
      (B-C-D) not closed, whereas (C-D-E) is.

    * the loop init has not been a loop front in a prior closed
      loop. For example F would close the loops (D-E-F) but D is
      already front of the closed loop (C-D-E).

    * the load level of the front has already been covered by a prior
      turning point. Otherwise it is considered part of the front
      residuum.

    When a loop is closed it is possible that the loop back also
    closes unclosed loops of the past by acting as loop back for an
    unclosed start/front pair. For example E closes the loop (C-D-E)
    and then also (A-B-E).

    ::

        Load -----------------------------
        |        x B               F x
        --------/-\-----------------/-----
        |      /   \   x D         /
        ------/-----\-/-\---------/-------
        |    /     C x   \       /
        --\-/-------------\-----/---------
        |  x A             \   /
        --------------------\-/-----------
        |                    x E
        ----------------------------------
        |              Time

    .. _subsection_TP: ../demos/rainflow.ipynb#Classic-Three-Point-Counting
    '''
    def __init__(self):
        super(RainflowCounterThreePoint, self).__init__()
        self._residuals = None

    @cython.locals(
        start=cython.int, front=cython.int, back=cython.int,
        highest_front=cython.int, lowest_front=cython.int,
        start_val=cython.double, front_val=cython.double, back_val=cython.double,
        turns=cython.double[:])
    def process(self, samples):
        ''' Processes a sample chunk

        Parameters
        ----------
        samples : array_like, shape (N, )
            The samples to be processed

        Returns
        -------
        self : RainflowCounterThreePoint
            The ``self`` object so that processing can be chained

        Example
        -------
        >>> rfc = RainflowCounterThreePoint().process(samples)
        >>> rfc.get_rainflow_matrix_frame(128)
        '''
        if self._residuals is None:
            residuals = samples[:1]
            residual_indeces = [0, 1]
        else:
            residuals = self._residuals[:-1]
            residual_indeces = [*range(len(residuals))]

        turns_np = np.concatenate((residuals, self._get_new_turns(samples), samples[-1:]))
        turns = turns_np

        highest_front = np.argmax(residuals)
        lowest_front = np.argmin(residuals)

        back = residual_indeces[-1] + 1
        while back < turns.shape[0]:
            if len(residual_indeces) >= 2:
                start = residual_indeces[-2]
                front = residual_indeces[-1]
                start_val, front_val, back_val = turns[start], turns[front], turns[back]

                if front_val > turns[highest_front]:
                    highest_front = front
                if front_val < turns[lowest_front]:
                    lowest_front = front

                if (start >= max(lowest_front, highest_front) and
                    np.abs(back_val - front_val) >= np.abs(front_val - start_val) and
                    front != highest_front and front != lowest_front):
                    self.loops_from.append(start_val)
                    self.loops_to.append(front_val)
                    residual_indeces.pop()
                    residual_indeces.pop()
                    continue

            residual_indeces.append(back)
            back += 1

        self._residuals = turns_np[residual_indeces]

        return self


class RainflowCounterFKM(AbstractRainflowCounter):
    '''Implements a rainflow counter as described in FKM non linear

    See the `here <subsection_FKM_>`_ in the demo for an example.

    The algorithm has been published by Clormann & Seeger 1985 and has
    been cited havily since.

    .. _subsection_FKM: ../demos/rainflow.ipynb#Algorithm-recommended-by-FKM-non-linear
    '''
    def __init__(self):
        super(RainflowCounterFKM, self).__init__()
        self._ir = 1
        self._residuals = []
        self._max_turn = 0.0

    @cython.locals(
        turns=cython.double[:],
        iz=cython.int, ir=cython.int,
        last0=cython.double, last1=cython.double,
        loop_assumed=cython.int,
        max_turn=cython.double)
    def process(self, samples):
        ''' Processes a sample chunk

        Parameters
        ----------
        samples : array_like, shape (N, )
            The samples to be processed

        Returns
        -------
        self : RainflowCounterFKM
            The ``self`` object so that processing can be chained

        Example
        -------
        >>> rfc = RainflowCounterFKM().process(samples)
        >>> rfc.get_rainflow_matrix_frame(128)
        '''
        ir = self._ir
        max_turn = self._max_turn
        turns = self._get_new_turns(samples)

        for current in turns:
            loop_assumed = True
            while loop_assumed:
                iz = len(self._residuals)
                loop_assumed = False
                if iz > ir:
                    last0 = self._residuals[-1]
                    last1 = self._residuals[-2]
                    if np.abs(current-last0) >= np.abs(last0-last1):
                        self.loops_from.append(last1)
                        self.loops_to.append(last0)
                        self._residuals.pop()
                        self._residuals.pop()
                        if np.abs(last0) < max_turn and np.abs(last1) < max_turn:
                            loop_assumed = True
                elif iz == ir:
                    if np.abs(current) > max_turn:
                        ir += 1
            max_turn = max(np.abs(current), max_turn)
            self._residuals.append(current)

            self._ir = ir
            self._max_turn = max_turn

        return self



# Copyright (c) 2020-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Gyöngyvér Kiss"
__maintainer__ = __author__

import numpy as np
import h5py
from h5py.h5t import string_dtype

from .exceptions import *
from .vmap_dataset import VMAPDataset


class VMAPElementType(VMAPDataset):
    def __init__(self, identifier, type_name, type_description, number_of_nodes, dimensions, shape_type, interpolation_type,
                 integration_type, number_of_normal_components, number_of_shear_components,
                 connectivity=None, face_connectivity=None):
        super().__init__(identifier)
        self._type_name = type_name
        self._type_description = type_description
        self._number_of_nodes = number_of_nodes
        self._dimension = dimensions
        self._shape_type = shape_type
        self._interpolation_type = interpolation_type
        self._integration_type = integration_type
        self._number_of_normal_components = number_of_normal_components
        self._number_of_shear_components = number_of_shear_components

        self._connectivity = []
        if connectivity is not None:
            self._connectivity = connectivity

        self._face_connectivity = []
        if face_connectivity is not None:
            self._face_connectivity = face_connectivity

    @property
    def attributes(self):
        if self._identifier is None:
            raise (APIUseError("Need to set_identifier() before requesting the attributes."))
        return (self._identifier, self._type_name, self._type_description, self._number_of_nodes, self._dimension,
                self._shape_type, self._interpolation_type, self._integration_type, self._number_of_normal_components,
                self._number_of_shear_components, np.array(self._connectivity), np.array(self._face_connectivity))

    @property
    def dtype(self):
        dt_type = np.dtype({"names": ["myIdentifier", "myTypeName", "myTypeDescription", "myNumberOfNodes",
                                      "myDimension", "myShapeType", "myInterpolationType", "myIntegrationType",
                                      "myNumberOfNormalComponents", "myNumberOfShearComponents", "myConnectivity",
                                      "myFaceConnectivity"],
                            "formats": ['<i4', string_dtype(), string_dtype(), '<i4', '<i4', '<i4', '<i4', '<i4',
                                        '<i4', '<i4', h5py.special_dtype(vlen=np.dtype('int32')),
                                        h5py.special_dtype(vlen=np.dtype('int32'))]})
        return dt_type

    @property
    def dataset_name(self):
        return 'ELEMENTTYPES'

    @property
    def type_name(self):
        return self._type_name



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import pandas as pd

from matplotlib.patches import Polygon
from matplotlib.collections import PatchCollection
from matplotlib import cm as colormap

from pylife.mesh import meshsignal


@pd.api.extensions.register_dataframe_accessor('meshplot')
class PlotmeshAccessor(meshsignal.MeshAccessor):
    ''' Plot a value on a 2d mesh

    The accessed DataFrame must be accessible by :class:`meshsignal.MeshAccessor`.
    '''
    def plot(self, axis, value_key, **kwargs):
        '''Plot the accessed dataframe

        Parameters
        ----------
        ax : matplotlib.pyplot.axis
            The axis to plot on

        value_key : str
            The column name in the accessed DataFrame

        **kwargs
            Arguments passed to the :class:`matplotlib.collections.PatchCollection`,
            like for example ``cmap``

        See also
        --------
        :func:`pandas.api.extensions.register_dataframe_accessor()`
        '''
        patches = []
        xmin = self._obj.x.min()
        xmax = self._obj.x.max()
        ymin = self._obj.y.min()
        ymax = self._obj.y.max()

        for _, el in self._obj.groupby('element_id'):
            nds = el.loc[el.index.get_level_values('node_id')[0:el.shape[0]//2], ['x', 'y']].to_numpy()
            patches.append(Polygon(nds, closed=True))

        sf = PatchCollection(patches, **kwargs)
        sf.set_array(self._obj[value_key].groupby('element_id').mean())

        axis.add_collection(sf)
        axis.set_xlim((xmin, xmax))
        axis.set_ylim((ymin, ymax))
        axis.figure.colorbar(sf, ax=axis)



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Small helper functions for fatigue analysis

"""

__author__ = "Cedric Philip Wagner"
__maintainer__ = "Johannes Mueller"

import numpy as np


def solidity_haibach(collective, k):
    """Compute solidity according to Haibach

    Refer to:
    Haibach - Betriebsfestigkeit - 3. Auflage (2005) - S.271

    Parameters
    ----------
    collective : np.ndarray
        numpy array of shape (:, 2) where ":" depends on the number of classes
        defined for the rainflow counting

            1. column: class values in ascending order

            2. column: accumulated number of cycles first entry is the total
               number of cycles then in a descending manner till the number of
               cycles of the highest stress class

    k : float
        slope of the S/N curve

    Returns
    -------
    V : np.ndarray (1,)
        Völligkeitswert (solidity)

    """

    S = collective[:, 0]
    # the accumulated number of cycles
    N_acc = collective[:, 1]

    # the number of cycles for each class
    hi = np.zeros_like(N_acc)
    # get the number of cycles for each class
    for i in range(len(hi)):
        if i == (len(hi) - 1):
            # the last entry is the accumulation of only the last class
            # so it is already the number of cycles of the highest class
            hi[i] = N_acc[i]
        else:
            hi[i] = N_acc[i] - N_acc[i + 1]

    # the selection of S is required so that the highest class
    # with actual counts (hi > 0) is taken as reference for all stress values
    xi = S / S[hi > 0].max()

    V = np.sum((hi * (xi**k)) / hi.sum())

    return V


def solidity_fkm(collective, k):
    """Compute solidity according to the FKM guideline (2012)

    Refer to:
    FKM-Richtlinie - 6. Auflage (2012) - S.58 - Gl. (2.4.55) +  Gl. (2.4.55)

    Parameters
    ----------
    collective : np.ndarray
        numpy array of shape (:, 2) where ":" depends on the number of classes
        defined for the rainflow counting

            1. column: class values in ascending order

            2. column: accumulated number of cycles first entry is the total
               number of cycles then in a descending manner till the number of
               cycles of the highest stress class k : float slope of the S/N
               curve

    Returns
    -------
    V : np.ndarray
        Völligkeitswert (solidity)

    """

    V_haibach = solidity_haibach(collective, k)
    V = V_haibach**(1./k)

    return V


class StressRelations:
    """Namespace for simple relations of stress / amplitude / R-ratio

    Refer to:
    Haibach (2006), p. 21
    """

    @staticmethod
    def get_max_stress_from_amplitude(amplitude, R):
        return 2 * amplitude / (1 - R)

    @staticmethod
    def get_mean_stress_from_amplitude(amplitude, R):
        return amplitude * (1 + R) / (1 - R)


def irregularity_factor(rainflow_matrix, residuals=np.empty(0), decision_bin=None):
    """
    Calculate the irregularity factor of a turning point sequence based on a rainflow matrix and its residuals.

    Two sided irregularity factor:

        ..math::

        I = N_{mean crossings} / N_{turning points}

    Parameters
    ----------
    rainflow_matrix: np.ndarray[int, int]
        2D-rainflow matrix (must be square shaped)
    residuals: np.ndarray[int], Optional
        1D array of residuals to consider for accurate calculation. Consecutive duplicates are removed beforehand.
        Residuals must be provided as bin numbers.
        Hint: Transformation from physical to binned values possible via np.digitize.
    decision_bin: int, Optional
        Bin number that equals the mean (two-sided). If not provided the decision_bin is inferred by the matrix entries
        as the mean value based on the turning points and will be broadcasted to int-type.

    Todo
    ----
    Future version may provide the one-sided irregularity factor as a second option. Formula would be:

    One sided irregularity factor:

        .. math::

        I = N_{zero bin upwards crossing} / N_{peaks}

    N_{zero bin upwards crossings} equals positive_mean_bin_crossing if `decision_bin` is set to the bin of physical 0.
    Inferring exact amount of peaks from rainflow-matrix and residuals is left to be done.
    """
    # Ensure input types
    assert isinstance(rainflow_matrix, np.ndarray)
    assert isinstance(residuals, np.ndarray)
    if rainflow_matrix.shape[0] != rainflow_matrix.shape[1]:
        raise ValueError("Rainflow matrix must be square shaped in order to calculate the irregularity factor.")

    # Remove duplicates from residuals
    diffs = np.diff(residuals)
    if np.any(diffs == 0.0):
        # Remove the duplicates
        duplicates = np.concatenate([diffs == 0, [False]])
        residuals = residuals[~duplicates]

    # Infer decision bin as mean if necessary
    if decision_bin is None:
        row_sum = 0
        col_sum = 0
        total_counts = 0
        for i in range(rainflow_matrix.shape[0]):
            row = rainflow_matrix[i, :].sum()
            col = rainflow_matrix[:, i].sum()
            total_counts += row + col
            row_sum += i * row
            col_sum += i * col

        total_counts += residuals.shape[0]
        res_sum = residuals.sum()

        decision_bin = int((row_sum + col_sum + res_sum) / total_counts)
    else:
        decision_bin = int(decision_bin)

    # Calculate two sided irregularity factor
    positive_mean_bin_crossing = rainflow_matrix[0:decision_bin, decision_bin:-1].sum()
    negative_mean_bin_crossing = rainflow_matrix[decision_bin:-1, 0:decision_bin].sum()
    total_mean_crossing = 2 * (positive_mean_bin_crossing + negative_mean_bin_crossing)
    amount_of_turning_points = 2 * rainflow_matrix.sum()

    amount_of_turning_points += residuals.shape[0]
    for i in range(residuals.shape[0] - 1):
        if (residuals[i] - decision_bin) * (residuals[i+1] - decision_bin) < 0:
            total_mean_crossing += 1
    return total_mean_crossing / amount_of_turning_points



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import inspect
import pandas as pd


class DataValidator:
    def get_missing_keys(self, signal, keys_to_check):
        '''Gets a list of missing keys that are needed for a signal object

        Parameters
        ----------
        signal : pandas.DataFrame or pandas.Series
            The object to be checked

        keys_to_check : list
            A list of keys that need to be available in `signal`

        Returns
        -------
        missing_keys : list
            a list of missing keys

        Raises
        ------
        AttributeError
            If `signal` is neither a `pandas.DataFrame` nor a `pandas.Series`

        Notes
        -----
        If `signal` is a `pandas.DataFrame`, all keys of
        `keys_to_check` not found in the `signal.columns` are
        returned.

        If `signal` is a `pandas.Series`, all keys of
        `keys_to_check` not found in the `signal.index` are
        returned.
        '''
        if isinstance(signal, pd.Series):
            keys_avail = signal.index
        elif isinstance(signal, pd.DataFrame):
            keys_avail = signal.columns
        else:
            raise AttributeError("An accessor object needs to be either a pandas.Series or a pandas.DataFrame")

        missing_keys = []
        for k in keys_to_check:
            if k not in keys_avail:
                missing_keys.append(k)
        return missing_keys


    def fail_if_key_missing(self, signal, keys_to_check, msg=None):
        '''Raises an exception if any key is missing in a signal object

        Parameters
        ----------
        signal : pandas.DataFrame or pandas.Series
            The object to be checked

        keys_to_check : list
            A list of keys that need to be available in `signal`

        Raises
        ------
        AttributeError
            if `signal` is neither a `pandas.DataFrame` nor a `pandas.Series`
        AttributeError
            if any of the keys is not found in the signal's keys.

        Notes
        -----
        If `signal` is a `pandas.DataFrame`, all keys of
        `keys_to_check` meed to be found in the `signal.columns`.

        If `signal` is a `pandas.Series`, all keys of
        `keys_to_check` meed to be found in the `signal.index`.

        See also
        --------
        :func:`signal.get_missing_keys`
        :class:`stresssignal.StressTensorVoigtAccessor`
        '''
        missing_keys = self.get_missing_keys(signal, keys_to_check)
        if not missing_keys:
            return
        if msg is None:
            stack = inspect.stack()
            the_class = stack[2][0].f_locals['self'].__class__
            msg = the_class.__name__ + ' must have the items %s. Missing %s.'
        raise AttributeError(msg % (', '.join(keys_to_check), ', '.join(missing_keys)))

    @staticmethod
    def fill_member(key, values):
        if key in values:
            member = values[key]
        else:
            member = ''
        return member



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Daniel Christopher Kreuter"
__maintainer__ = "Johannes Mueller"

import pandas as pd
import pylife.mesh.meshsignal as meshsignal


@pd.api.extensions.register_dataframe_accessor('hotspot')
class HotSpot(meshsignal.MeshAccessor):

    def calc(self, value_key, limit_frac=0.9, artefact_threshold=None):
        '''Calculates hotspots on a FE mesh

        Parameters
        ----------
        value_key : string
            Column name of the field variable, on which the Hot Spot
            calculation is done.

        limit_frac : float, optional
            Fraction of the max field variable. Example: If you set
            limit_frac = 0.9, the function finds all nodes and regions
            which are >= 90% of the maximum value of the field
            variable.  default: 0.9

        artefact_threshold : float, optional
            If set all the values above the `artefact_threshold` limit are not
            taken into account for the calculation of the maximum value. This
            is meant to be used for numerical artefacts which would take
            the threshold value for hotspot determined by `limit_frac` to such
            a high level, that all the relevant hotspots would "hide" underneath
            it.

        Returns
        -------
        hotspots : pandas.Series
            A Series of intergers with the same index of the accessed
            mesh object indicating which mesh point belongs to which hotspot.
            A value 0 means below the `limit_frac`.

        Notes
        -----
        A loop is defined in the following way:
                * Select the node with the maximum stress value
                * Find all elements > `limit_frac` belonging to this node
                * Select all nodes > `limit_frac` belonging to these elements
                * Start loop again until all nodes > `limit_frac` are assigned to a hotspot

        Attention: All stress values are node based, not integration point based
        '''
        max_value = (self._obj[value_key].max() if artefact_threshold is None
                     else self._obj.loc[self._obj[value_key] < artefact_threshold, value_key].max())
        above_limit = self._obj[value_key] >= limit_frac*max_value
        hotspots = pd.Series(0, name='hotspot', index=self._obj.index)

        hs_index = 1
        while above_limit.any():
            hs = self.__hs_sel(above_limit, value_key)
            hotspots.loc[hs] = hs_index
            hs_index += 1
            above_limit ^= hs

        return hotspots

    def __hs_sel(self, remaining, value_key):
        max_index = self._obj.loc[remaining, value_key].idxmax()
        new_hotspot = pd.Series(False, self._obj.index)
        new_hotspot[max_index] = True

        sl = pd.IndexSlice

        new_entries = True
        while new_entries:
            new_entries = False
            new_nodes_idx = remaining[new_hotspot].index.get_level_values('node_id')
            new_elems_idx = remaining[new_hotspot].index.get_level_values('element_id')
            new_nodes = remaining.loc[sl[new_nodes_idx, :]] ^ new_hotspot
            new_elems = remaining.loc[sl[:, new_elems_idx]] ^ new_hotspot
            if new_nodes.any():
                new_entries = True
                new_hotspot[new_nodes] = True
            if new_elems.any():
                new_entries = True
                new_hotspot[new_elems] = True

        return new_hotspot



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Mustapha Kassem"
__maintainer__ = "Johannes Mueller"

import numpy as np
from scipy import stats

from pylife.utils.functions import scatteringRange2std


class WoehlerCurveGraph:

    def __init__(self, woehler_curve, y_min, y_max, failure_probablilty=0.5):
        self._woehler_curve = woehler_curve
        self.__calc_woehler_curve_points(y_min, y_max)
        self.__shift_woehlercurve_points(failure_probablilty)

    def __calc_woehler_curve_points(self, y_min, y_max):
        """ Basquin curve equation

        http://www.ux.uis.no/~hirpa/6KdB/ME/S-N%20diagram.pdf
        """
        if self._woehler_curve['k_1'] == np.inf:
            self._points = np.array([[self._woehler_curve['SD_50'], 1E9]])
        else:
            y = np.linspace(y_max, y_min, num=100)
            x = self._woehler_curve['ND_50']*(y/self._woehler_curve['SD_50'])**(-self._woehler_curve['k_1'])
            self._points = np.array([y, x]).transpose()

    def __shift_woehlercurve_points(self, pa_goal):
        """ Shift the Basquin-curve according to the failure probability value (obtain the 10-90 % curves)"""
        TN_inv = self._woehler_curve['1/TN']
        self._points[:, 1] /= 10**(-stats.norm.ppf(pa_goal)*scatteringRange2std(TN_inv))

    @property
    def points(self):
        return self._points



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Daniel Christopher Kreuter"
__maintainer__ = "Johannes Mueller"


import numpy as np
import pandas as pd


def combine_hist(hist_list, method='sum', nbins=64):
    """
    Performs the combination of multiple Histograms.

    Parameters
    ----------

    hist_list: list
        list of histograms with all histograms (saved as DataFrames in pyLife format)
    method: str
        method: 'sum', 'min', 'max', 'mean', 'std'  default is 'sum'
    nbins: int
        number of bins of the combined histogram

    Returns
    -------

    DataFrame:
        Combined histogram
    list:
        list with the reindexed input histograms

    """

    hist_combined = pd.concat(hist_list)
    index_min = hist_combined.index.left.min()
    index_max = hist_combined.index.right.max()

    kwargs = {'ddof': 0} if method == 'std' else {}
    return (hist_combined
            .groupby(pd.cut(hist_combined.index.mid.values, np.linspace(index_min, index_max, nbins+1)))
            .agg(method, **kwargs)
            .set_index(pd.interval_range(index_min, index_max, nbins, name='range')))



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import numpy as np
import pandas as pd
import scipy.stats as stats
import scipy.signal as signal

class TimeSignalGenerator:
    '''Generates mixed time signals

    The generated time signal is a mixture of random sets of

    * sinus signals
    * gauss signals (not yet)
    * log gauss signals (not yet)

    For each set the user supplys a dict describing the set::

      sinus_set = {
          'number': number of signals
          'amplitude_median':
          'amplitude_std_dev':
          'frequency_median':
          'frequency_std_dev':
          'offset_median':
          'offset_std_dev':
      }

    The amplitudes (:math:`A`), fequencies (:math:`\omega`) and
    offsets (:math:`c`) are then norm distributed. Each sinus signal
    looks like

            :math:`s = A \sin(\omega t + \phi) + c`

    where :math:`phi` is a random value between 0 and :math:`2\pi`.

    So the whole sinus :math:`S` set is given by the following expression:

            :math:`S = \sum^n_i A_i \sin(\omega_i t + \phi_i) + c_i`.
    '''

    def __init__(self, sample_rate, sine_set, gauss_set, log_gauss_set):
        sine_amplitudes = stats.norm.rvs(loc=sine_set['amplitude_median'],
                                         scale=sine_set['amplitude_std_dev'],
                                         size=sine_set['number'])
        sine_frequencies = stats.norm.rvs(loc=sine_set['frequency_median'],
                                          scale=sine_set['frequency_std_dev'],
                                          size=sine_set['number'])
        sine_offsets = stats.norm.rvs(loc=sine_set['offset_median'],
                                      scale=sine_set['offset_std_dev'],
                                      size=sine_set['number'])
        sine_phases = 2. * np.pi * np.random.rand(sine_set['number'])

        self.sine_set = list(zip(sine_amplitudes, sine_frequencies, sine_phases, sine_offsets))

        self.sample_rate = sample_rate
        self.time_position = 0.0

    def query(self, sample_num):
        '''Gets a sample chunk of the time signal

        Parameters
        ----------
        sample_num : int
            number of the samples requested

        Returns
        -------
        samples : 1D numpy.ndarray
            the requested samples


        You can query multiple times, the newly delivered samples
        will smoothly attach to the previously queried ones.
        '''
        samples = np.zeros(sample_num)
        end_time_position = self.time_position + (sample_num-1) / self.sample_rate

        for ampl, omega, phi, offset in self.sine_set:
            periods = np.floor(self.time_position / omega)
            start = self.time_position - periods * omega
            end = end_time_position - periods * omega
            time = np.linspace(start, end, sample_num)
            samples += ampl * np.sin(omega * time + phi) + offset

        self.time_position = end_time_position + 1. / self.sample_rate

        return samples

    def reset(self):
        ''' Resets the generator

        A resetted generator behaves like a new generator.
        '''
        self.time_position = 0.0


class TimeSignalPrep:

    def __init__(self,df):

        self.df = df

    def resample_acc(self,sample_rate_new = 1):
        """ Resampling the time series

        Parameters
        ----------
        self: DataFrame

        time_col: str
            column name of the time column
        sample_rate_new: float
            sample rate of the resampled time series

        Returns
        -------
        DataFrame
        """
#        dfResample.index =  np.arange(self.df.index.min(),self.df.index.max(),1/sample_rate_new)
        index_new =  np.linspace(self.df.index.min(),
                                 self.df.index.min() + np.floor((self.df.index.max()-self.df.index.min())*sample_rate_new)/sample_rate_new,
                                 int(np.floor(self.df.index.max()-self.df.index.min())*sample_rate_new + 1))
        dfResample = pd.DataFrame(index = index_new)
        for colakt in self.df.columns:
            dfResample[colakt] = np.interp(dfResample.index,self.df.index,self.df[colakt])
        return dfResample

    def butter_bandpass(self, lowcut, highcut, fs, order=5):
        """Use the functonality of scipy"""
        nyq = 0.5 * fs
        low = lowcut / nyq
        high = highcut / nyq
        b, a = signal.butter(order, [low, high], btype='band')
        TSout = signal.filtfilt(b, a, self.df)
        return TSout

    def running_stats_filt(self,col,window_length = 2048,buffer_overlap = 0.1,limit = 0.05, method = "rms"):
        """
        Calculates the running statistics of one DataFrame column and drops the rejected data points from the whole DataFrame.

        **Attention**: Reset_index is used

        Parameters
        -----------

        self: DataFrame

        col: str
            column name of the signal for the runnings stats calculation
        window_length: int
            window length of the single time snippet, default is 2048
        buffer_overlap: float
            overlap parameter, 0.1 is equal to 10 % overlap of every buffer, default is 0.1
        limit: float
            limit value of skipping values, 0.05 for example skips all values which buffer method parameter is lower than 5% of the total max value,
            default is 0.05
        method: str
            method: 'rms', 'min', 'max', 'abs', default is 'rms'

        Returns
        -------
        DataFrame

        """
        df = self.df.reset_index(drop = True)
        delta_t = self.df.index.values[1]-self.df.index.values[0]
        hop = int(window_length*(1-buffer_overlap)) # absolute stepsize
        df = df.loc[:int(np.floor(len(df)/hop)*hop),:]
        n_iter = 1+int((len(df)-window_length)/(hop))
        ind_act = 0
        stats_list = []
        for ii in range (n_iter):
            if method == "rms":
                stats_list.append( np.sqrt(np.mean(df[col][ind_act:ind_act+window_length]**2)))
            elif method == "max":
                stats_list.append(np.max(df[col][ind_act:ind_act+window_length]))
            elif method == "min":
                stats_list.append(np.abs(np.min(df[col][ind_act:ind_act+window_length])))
            elif method == "abs":
                stats_list.append(np.max(np.abs(df[col][ind_act:ind_act+window_length])))
            ind_act = ind_act+hop
        try:
            stats_list = pd.DataFrame({"stats": np.asarray(stats_list)})#,
        except:
            print(str(stats_list))
                                  # index = np.arange(0,len(np.asarray(stats_list))-1,
                                  #                   np.asarray(stats_list)))
        stats_list = stats_list[stats_list["stats"] < limit*stats_list["stats"].max()]
        for ind_act in stats_list.index:
            df = df.drop(index = np.arange(ind_act*hop,ind_act*hop+window_length), errors = 'ignore')
        df.index = np.linspace(0,delta_t*(len(df)-1), len(df))
        return df



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .elementary import Elementary

import pandas as pd
import numpy as np
from scipy import optimize
import mystic as my
import warnings

class MaxLikeInf(Elementary):
    def _specific_analysis(self, wc):
        SD_50, TS_inv = self.__max_likelihood_inf_limit()

        wc['SD_50'] = SD_50
        wc['1/TS'] = TS_inv
        wc['ND_50'] = self._transition_cycles(SD_50)

        return wc

    def __max_likelihood_inf_limit(self):
        ''' This maximum likelihood procedure estimates the load endurance limit SD50_mali_2_param and the
        scatter in load direction TS_mali_2_param.
        Moreover, the load cycle endurance is computed by the interesecting endurance limit line with the
        line of slope k_1
        '''
        SD_start = self._fd.fatigue_limit
        TS_start = 1.2

        var_opt = optimize.fmin(lambda p: -self._lh.likelihood_infinite(p[0], p[1]),
                                [SD_start, TS_start], disp=False, full_output=True)

        SD_50 = var_opt[0][0]
        TS_inv = var_opt[0][1]

        return SD_50, TS_inv


class MaxLikeFull(Elementary):
    def _specific_analysis(self, wc, fixed_parameters={}):
        return pd.Series(self.__max_likelihood_full(wc, fixed_parameters))

    def __max_likelihood_full(self, initial_wcurve, fixed_prms):
        """
        Maximum likelihood is a method of estimating the parameters of a distribution model by maximizing
        a likelihood function, so that under the assumed statistical model the observed data is most probable.
        This procedure consists of estimating the  curve parameters, where some of these paramters may
        be fixed by the user. The remaining paramters are then fitted to produce the best possible outcome.
        The procedure uses the function Optimize.fmin
        Optimize.fmin iterates over the function mali_sum_lolli values till it finds the minimum

        https://en.wikipedia.org/wiki/Maximum_likelihood_estimation

        Parameters
        ----------
        self.p_opt: Start values of the Mali estimated parameters if none are fixed by the user.

        self.dict_bound: Boundary values of the Mali estimated parameters if none are fixed by the user.
        This forces the optimizer to search for a minimum solution within a given area.


        Returns
        -------
        self.Mali_5p_result: The estimated parameters computed using the optimizer.

        """
        def warn_and_fix_if_no_runouts():
            nonlocal fixed_prms
            if self._fd.num_runouts == 0:
                warnings.warn(UserWarning("MaxLikeHood: no runouts are present in fatigue data. "
                                          "Proceeding with SD_50 = 0 and 1/TS = 1 as fixed parameters. "
                                          "This is NOT a standard evaluation!"))
                fixed_prms = fixed_prms.copy()
                fixed_prms.update({'SD_50': 0.0, '1/TS': 1.0})

        def fail_if_less_than_three_fractures():
            if self._fd.num_fractures < 3 or len(self._fd.fractured_loads) < 2:
                raise ValueError("MaxLikeHood: need at least three fractures on two load levels.")

        def warn_and_fix_if_less_than_two_mixed_levels():
            nonlocal fixed_prms
            if len(self._fd.mixed_loads) < 2 and self._fd.num_runouts > 0 and self._fd.num_fractures > 0:
                warnings.warn(UserWarning("MaxLikeHood: less than two mixed load levels in fatigue data."
                                          "Proceeding by setting a predetermined scatter from the standard Wöhler curve."))
                fixed_prms = fixed_prms.copy()
                TN_inv, TS_inv = self._pearl_chain_method()
                fixed_prms.update({'1/TS': TS_inv})

        fail_if_less_than_three_fractures()
        warn_and_fix_if_no_runouts()
        warn_and_fix_if_less_than_two_mixed_levels()

        p_opt = initial_wcurve.to_dict()
        for k in fixed_prms:
            p_opt.pop(k)

        if not p_opt:
            raise AttributeError('You need to leave at least one parameter empty!')
        var_opt = my.scipy_optimize.fmin(
            self.__likelihood_wrapper, [*p_opt.values()],
            args=([*p_opt], fixed_prms),
            full_output=True,
            disp=False,
            maxiter=1e4,
            maxfun=1e4,
        )
        res = {}
        res.update(fixed_prms)
        res.update(zip([*p_opt], var_opt[0]))

        return self.__make_parameters(res)

    def __make_parameters(self, params):
        params['SD_50'] = np.abs(params['SD_50'])
        params['1/TS'] = np.abs(params['1/TS'])
        params['k_1'] = np.abs(params['k_1'])
        params['ND_50'] = np.abs(params['ND_50'])
        params['1/TN'] = np.abs(params['1/TN'])
        return params

    def __likelihood_wrapper(self, var_args, var_keys, fix_args):
        ''' 1) Finds the start values to be optimized. The rest of the paramters are fixed by the user.
            2) Calls function mali_sum_lolli to calculate the maximum likelihood of the current
            variable states.
        '''
        args = {}
        args.update(fix_args)
        args.update(zip(var_keys, var_args))
        args = self.__make_parameters(args)

        return -self._lh.likelihood_total(args['SD_50'], args['1/TS'], args['k_1'], args['ND_50'], args['1/TN'])



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Mesh Mapping
============

Map values of one FEM mesh into another

'''

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import scipy.interpolate as interp
import numpy as np
import pandas as pd
from pylife.mesh.meshsignal import PlainMeshAccessor

@pd.api.extensions.register_dataframe_accessor('meshmapper')
class MeshmapperAccessor(PlainMeshAccessor):
    '''Mapper to map points of one mesh to another

    Notes
    -----
    The accessed DataFrame needs to be accessible by a :class:`PlainMeshAccessor`.
    '''
    def process(self, from_df, value_key, method='linear'):
        '''Performs the mapping

        Parameters
        ----------
        from_df : pandas.DataFrame accessible by a :class:`PlainMeshAccessor`.
            The DataFrame that is to be mapped to the accessed one.
            Needs to have the same dimensions (2D or 3D) as the accessed one
        '''
        crd = self._coord_keys
        from_df.plain_mesh
        newvals = interp.griddata(from_df[crd], from_df[value_key], self._obj[crd], method=method)

        return pd.DataFrame({value_key: newvals}).set_index(self._obj.index)



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Simone Schreijäg"
__maintainer__ = "Johannes Mueller"

import numpy as np


def true_strain(tech_strain):
    """
    Calculation of true strain data (from experimental data generated by tensile experiments)

    Parameters
    ----------
    tech_strain: array-like float

    Returns
    -------
    true_strain: array-like float

   """
    return np.log(1. + tech_strain)


def true_stress(tech_stress, tech_strain):
    """
     Calculate the true stress data from technical data

    Parameters
    ----------
    tech_stress : array-like float
        stress data from tensile experiments
    tech_strain : list of float
        strain data from tensile experiments

    Returns
    -------
    true_stress : array-like float

    """
    return tech_stress * (1. + tech_strain)


def true_fracture_strain(reduction_area_fracture):
    """
    Calculation of the true frature strain (in the FKM Non Linear (static assessment))
    Parameters
    ----------
    reduction_area_fracture : float
        directly measured on the fractures sample

    Returns
    -------
    true_fracture_strain: float
        descrivbes the calculated true fracture strain.

    """
    return np.log(1./(1. - reduction_area_fracture))


def true_fracture_stress(fracture_force, initial_cross_section, reduction_area_fracture):
    """
    Calculation of the true fracture stress (euqation FKM Non-linear (static assessment))

    Parameters
    ----------
    fracture_force : float
        from experimental results
    initial_cross_section : float
        cross section of initial tensile sample.
    reduction_area_fracture : float
         directly measured on the fractures sample.

    Returns
    -------
    true_fracture_stress: float
        calculated true fracture stress of the sample.

    """
    return fracture_force/(initial_cross_section * (1. - reduction_area_fracture))



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""

Implementation of the miner rule for fatigue analysis
=====================================================

Currently, the following implementations are part of this module:

* Miner-elementar
* Miner-haibach

The source will be given in the function/class

References
----------
M. Wächter, C. Müller and A. Esderts, "Angewandter Festigkeitsnachweis nach {FKM}-Richtlinie"
Springer Fachmedien Wiesbaden 2017, https://doi.org/10.1007/978-3-658-17459-0

E. Haibach, "Betriebsfestigkeit", Springer-Verlag 2006, https://doi.org/10.1007/3-540-29364-7
"""

__author__ = "Cedric Philip Wagner"
__maintainer__ = "Johannes Mueller"


import logging
import numpy as np
import pandas as pd

from pylife.strength.helpers import solidity_haibach
from pylife.strength.sn_curve import FiniteLifeCurve

DEBUG = False
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
ch = logging.StreamHandler()
formatter = logging.Formatter(
    '%(asctime)s - %(name)s - %(levelname)s - %(message)s\n')
ch.setFormatter(formatter)
if DEBUG:
    ch.setLevel(logging.DEBUG)
else:
    ch.setLevel(logging.INFO)
logger.addHandler(ch)


def get_accumulated_from_relative_collective(collective):
    """Get collective with accumulated frequencies

    This function can be used to transform a collective with
    relative frequencies.

    Parameters
    ----------
    collective : np.ndarray
        numpy array of shape (:, 2)
        where ":" depends on the number of classes defined
        for the rainflow counting
        * column: class values in ascending order
        * column: relative number of cycles for each load class
    """
    accumulated = np.stack([collective[:, 0],
                            np.flipud(np.cumsum(np.flipud(collective[:, 1])))],
                           axis=1)
    return accumulated


class MinerBase:
    """Basic functions related to miner-rule (original)

    Definitions will be based on the given references.
    Therefore, the original names are used so that they can
    be looked up easily.

    Parameters
    ----------
    ND_50 : float
        number of cycles of the fatigue strength of the S/N curve [number of cycles]
    k_1 : float
        slope of the S/N curve [unitless]
    SD_50 : float
        fatigue strength of the S/N curve [MPa]
    """

    collective = None

    def __init__(self, ND_50, k_1, SD_50):
        self.ND_50 = abs(ND_50)
        if self.ND_50 < 10**3:
            raise ValueError(
                "ND_50 is unexpectedly small ('{}'). "
                "Please check valid input of this parameter! "
                "Input is not in logarithmic scale.".format(
                    self.ND_50,
                )
            )
        self.k_1 = abs(k_1)
        self.SD_50 = abs(SD_50)
        self.sn_curve = FiniteLifeCurve(
            k_1=self.k_1,
            SD_50=self.SD_50,
            ND_50=self.ND_50
        )

    def setup(self, collective):
        """Calculations independent from the instantation

        Use the setup for functions that might require information that was not
        yet available at runtime during instantation.

        Parameters
        ----------
        collective : np.ndarray
            numpy array of shape (:, 2)
            where ":" depends on the number of classes defined
            for the rainflow counting
            * column: class values in ascending order
            * column: accumulated number of cycles
            first entry is the total number of cycles
            then in a descending manner till the
            number of cycles of the highest stress class
        """
        self._parse_collective(collective)
        self.zeitfestigkeitsfaktor_collective = self.calc_zeitfestigkeitsfaktor(
            self.H0,
            total_lifetime=False,
        )

    def _parse_collective(self, collective):
        """Parse collective and structure frequently used features"""
        # first, only select values of the collective with number of cycles >0
        self.original_collective = collective.copy()
        if isinstance(collective, pd.DataFrame) \
                and isinstance(collective.index, pd.IntervalIndex):
            collective = self._transform_pylife_collective(collective)
            logger.debug("Recognized pylife-like transformed collective.")

        N_collective_accumulated = collective[:, 1]

        def get_relative_from_accumulated(accumulated):
            relative = np.append(
                np.abs(np.diff(accumulated)),
                # the last value has to be appended separately since
                # it is the only value that is not obtained as a difference
                accumulated[-1],
            )
            return relative

        hi = get_relative_from_accumulated(N_collective_accumulated)
        self.collective = collective[hi > 0]
        self._skipped_collective_values = collective[hi == 0]

        # normalize collective to S_a_max = 1
        self.collective = self.collective / np.array([self.collective[:, 0].max(), 1])

        # the first entry of 2nd column is the sum of all cycles
        self.H0 = self.collective[0, 1]

        self.S_collective = self.collective[:, 0]
        if not np.all(np.diff(self.S_collective) > 0):
            raise ValueError(
                "The load classes of the collective are not in ascending order: "
                "{}".format(self.S_collective)
            )
        self.N_collective_accumulated = self.collective[:, 1]
        # get the number of cycles for each class
        hi = get_relative_from_accumulated(self.N_collective_accumulated)
        self.N_collective_relative = hi
        self.collective_relative = np.stack((self.S_collective,
                                             self.N_collective_relative),
                                            axis=1)

    def _transform_pylife_collective(self, coll):
        """Adjust the mean stress corrected pylife collective for parsing

        Parameters
        ----------
        coll : pandas.DataFrame
            mean stress transformed pylife collective
        columns: frequency
        index: pandas.IntervalIndex
        """
        load_classes = np.array(coll.index.mid)
        self._original_pylife_frequencies = coll["frequency"].to_numpy()
        accumulated_frequencies = np.flip(
            np.cumsum(np.flip(self._original_pylife_frequencies))
        )

        return np.stack([load_classes, accumulated_frequencies], axis=1)

    def calc_zeitfestigkeitsfaktor(self, N, total_lifetime=True):
        """Calculate "Zeitfestigkeitsfaktor" according to Waechter2017 (p. 96)"""

        z = (self.ND_50 / N)**(1. / self.k_1)

        if total_lifetime:
            self.zeitfestigkeitsfaktor = z

        return z

    def calc_A(self, collective):
        """Compute multiple of the lifetime"""
        if collective is None:
            if self.collective is None:
                raise RuntimeError(
                    "The collective has not been specified either "
                    "as input parameter or as attribute during setup."
                )
        else:
            self._parse_collective(collective)

        if self.__class__.__name__ == "MinerBase":
            raise NotImplementedError(
                "Method should be used only by deriving classes."
            )

    def effective_damage_sum(self, A):
        """Compute 'effective damage sum' D_m

        Refers to the formula given in Waechter2017, p. 99

        Parameters
        ----------
        A : float or np.ndarray (with 1 element)
            the multiple of the lifetime
        """

        d_min = 0.3  # minimum as suggested by FKM
        d_max = 1.0

        d_m_no_limits = float(2. / (A**(1./4)))
        d_m = min(
            max(d_min, d_m_no_limits),
            d_max
        )
        self.d_m_no_limits = d_m_no_limits
        self.d_m = d_m

        return self.d_m

    def N_predict(self, load_level, A=None,):
        """The predicted lifetime according to damage sum of the collective

        Parameters
        ----------
        load_level : float
            the maximum (stress) amplitude of the collective
        A : float
            the lifetime multiple A
            BEWARE: this relation is only valid in a specific
            representation of the predicted (Miner) lifetime
            where the sn-curve is expressed via the point
            of the maximum amplitude of the collective:
            N_predicted = N(S = S_max) * A
        """
        # ignore limits
        #   e.g. for miner-elementary damage can be accumulated even with S_a,max below S_D
        #   methods like miner-haibach who do not allow that have internal sanity checks
        n_woehler_load_level = self.sn_curve.calc_N(load_level, ignore_limits=True)
        if A is None:
            A = self.calc_A()
        return n_woehler_load_level * A


class MinerElementar(MinerBase):
    """Implementation of Miner-elementar according to Waechter2017

    """
    # Solidity (Völligkeit) according to Haibach
    V_haibach = None
    # Solidity (Völligkeit) according to FKM guideline
    V_FKM = None

    def __init__(self, ND_50, k_1, SD_50):
        super(MinerElementar, self).__init__(
            ND_50, k_1, SD_50,
        )

    def setup(self, collective):
        super(MinerElementar, self).setup(collective)
        A = self.calc_A(self.collective)
        # at this point N_expected is equal to H0
        # i.e. the number of cycles for the collective (not N for sample failure)
        self.d_m_collective = self.effective_damage_sum(A)

    def calc_A(self, collective=None):
        """Compute the lifetime multiple according to miner-elementar

        Described in Waechter2017 as "Lebensdauervielfaches, A_ele".

        Parameters
        ----------
        collective : np.ndarray
            numpy array of shape (:, 2)
            where ":" depends on the number of classes defined
            for the rainflow counting
            * column: class values in ascending order
            * column: accumulated number of cycles
            first entry is the total number of cycles
            then in a descending manner till the
            number of cycles of the highest stress class
        """
        super(MinerElementar, self).calc_A(collective)
        V = solidity_haibach(self.collective, self.k_1)
        self.V_haibach = V
        self.V_FKM = V**(1/self.k_1)
        A = 1. / V
        self.A = A

        return A


class MinerHaibach(MinerBase):
    """Miner-modified according to Haibach (2006)

    WARNING: Contrary to Miner-elementar, the lifetime multiple A
             is not constant but dependent on the evaluated load level!

    Parameters
    ----------
    see MinerBase

    Attributes
    ----------
    A : dict
        the multiple of the life time initiated as dict
        Since A is different for each load level, the
        load level is taken as dict key (values are rounded to 0 decimals)
    """
    evaluated_load_levels = None

    def __init__(self, ND_50, k_1, SD_50):
        super(MinerHaibach, self).__init__(
            ND_50, k_1, SD_50,
        )
        self.evaluated_load_levels = {}

    def setup(self, collective):
        super(MinerHaibach, self).setup(collective)

    def calc_A(self, load_level, collective=None, ignore_inf_rule=False):
        """Compute the lifetime multiple for Miner-modified according to Haibach

        Refer to Haibach (2006), p. 291 (3.21-61). The lifetime multiple can be
        expressed in respect to the maximum amplitude so that
        N_lifetime = N_Smax * A

        Parameters
        ----------
        load_level : float > 0
            load level in [MPa]
        collective : np.ndarray (optional)
            the collective can optionally be input to this function
            if it is not specified, then the attribute is used.
            If no collective exists as attribute (is set during setup)
            then an error is thrown
        ignore_inf_rule : boolean
            By default, the lifetime is returned as inf when the given load level
            is smaller than the lifetime (see Haibach eq. 3.2-62).
            This rule can be ignored if an estimate for the lifetime in the
            region below the fatigue strength is required for investigation.

        Returns
        -------
        A : float > 0
            lifetime multiple
            return value is 'inf' if load_level < SD_50
        """
        super(MinerHaibach, self).calc_A(collective)
        self.evaluated_load_levels[round(load_level)] = {}
        # this parameter makes each evaluation of A unique
        self._last_load_level_evaluated = load_level
        if load_level < self.SD_50:
            logger.warning(
                "The given load level ('{}') is below SD_50 ('{}'). It is assumed "
                "that with this load level no damage occurs. "
                "Hence, an infinite value is returned.".format(
                    load_level,
                    self.SD_50
                )
            )
            if ignore_inf_rule:
                logger.warning("The rule to set the lifetime to infinity when "
                               "the given load level is lower than the fatigue "
                               "limit of the sn-curve has explicitly been ignored.")
            else:
                return np.inf
        assert self.S_collective.max() == 1
        s_a = self.S_collective * load_level
        i_full_damage = (s_a >= self.SD_50)
        i_reduced_damage = (s_a < self.SD_50)
        self.evaluated_load_levels[round(load_level)]["i_full_damage"] = i_full_damage
        self.evaluated_load_levels[round(load_level)]["i_reduced_damage"] = i_reduced_damage
        x_D = self.SD_50 / s_a.max()
        self.evaluated_load_levels[round(load_level)]["x_D"] = x_D

        s_full_damage = s_a[i_full_damage]
        s_reduced_damage = s_a[i_reduced_damage]
        n_full_damage = self.N_collective_relative[i_full_damage]
        n_reduced_damage = self.N_collective_relative[i_reduced_damage]

        # first expression of the summation term in the denominator
        sum_1 = np.dot(
            n_full_damage,
            ((s_full_damage / s_a.max())**self.k_1),
        )
        self.evaluated_load_levels[round(load_level)]["sum_1_denominator"] = sum_1
        sum_2 = (x_D**(1 - self.k_1)) * np.dot(
            n_reduced_damage,
            ((s_reduced_damage / s_a.max())**(2 * self.k_1 - 1))
        )
        self.evaluated_load_levels[round(load_level)]["sum_2_denominator"] = sum_2

        A = self.H0 / (sum_1 + sum_2)
        self.evaluated_load_levels[round(load_level)]["A"] = A

        return A

    def N_predict(self, load_level, A=None, ignore_inf_rule=False):
        """The predicted lifetime according to damage sum of the collective

        Parameters
        ----------
        load_level : float
            the maximum (stress) amplitude of the collective
        A : float
            the lifetime multiple A
            BEWARE: this relation is only valid in a specific
            representation of the predicted (Miner) lifetime
            where the sn-curve is expressed via the point
            of the maximum amplitude of the collective:
            N_predicted = N(S = S_max) * A
        """
        if A is None:
            A = self.calc_A(load_level, ignore_inf_rule=ignore_inf_rule)

        N_pred = super(MinerHaibach, self).N_predict(load_level, A)
        return N_pred



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
'''
Helper to process mesh based data
=================================

Data that is distributed over a geometrical body, e.g. a stress tensor
distribution on a component, is usually transported via a mesh. The
meshes are a list of items (e.g. nodes or elements of a FEM mesh),
each being described by the geometrical coordinates and the local data
values, like for example the local stress tensor data.

In a plain mesh (see :class:`PlainMeshAccessor`) there is no further
relation between the items is known, whereas a complete FEM mesh (see
:class:`MeshAccessor`) there is also information on the connectivity
of the nodes and elements.

Examples
--------
Read in a mesh from a vmap file:


>>> df = (vm = pylife.vmap.VMAPImport('demos/plate_with_hole.vmap')
             .make_mesh('1', 'STATE-2')
             .join_variable('STRESS_CAUCHY')
             .join_variable('DISPLACEMENT')
             .to_frame())
>>> df.head()
                            x         y    z        S11       S22  S33        S12  S13  S23        dx        dy   dz
element_id node_id
1          1734     14.897208  5.269875  0.0  27.080811  6.927080  0.0 -13.687358  0.0  0.0  0.005345  0.000015  0.0
           1582     14.555333  5.355806  0.0  28.319006  1.178649  0.0 -10.732705  0.0  0.0  0.005285  0.000003  0.0
           1596     14.630658  4.908741  0.0  47.701195  5.512213  0.0 -17.866833  0.0  0.0  0.005376  0.000019  0.0
           4923     14.726271  5.312840  0.0  27.699907  4.052865  0.0 -12.210032  0.0  0.0  0.005315  0.000009  0.0
           4924     14.592996  5.132274  0.0  38.010101  3.345431  0.0 -14.299768  0.0  0.0  0.005326  0.000013  0.0

Get the coordinates of the mesh.

>>> df.plain_mesh.coordinates.head()
                            x         y    z
element_id node_id
1          1734     14.897208  5.269875  0.0
           1582     14.555333  5.355806  0.0
           1596     14.630658  4.908741  0.0
           4923     14.726271  5.312840  0.0
           4924     14.592996  5.132274  0.0

Now the same with a 2D mesh:

>>> df.drop(columns=['z']).plain_mesh.coordinates.head()
                            x         y
element_id node_id
1          1734     14.897208  5.269875
           1582     14.555333  5.355806
           1596     14.630658  4.908741
           4923     14.726271  5.312840
           4924     14.592996  5.132274
'''

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import pandas as pd
from pylife import signal


@pd.api.extensions.register_dataframe_accessor("plain_mesh")
class PlainMeshAccessor(signal.PylifeSignal):
    '''DataFrame accessor to access plain 2D and 3D mesh data, i.e. without connectivity

    Raises
    ------
    AttributeError
        if at least one of the columns `x`, `y` is missing

    Notes
    -----
    The PlainMeshAccessor describes meshes whose only geometrical
    information is the coordinates of the nodes or elements. Unlike
    :class:`MeshAccessor` they don't know about connectivity, not even
    about elements and nodes.

    See also
    --------
    :class:`MeshAccessor`: accesses meshes with connectivity information
    :func:`pandas.api.extensions.register_dataframe_accessor()`: concept of DataFrame accessors
    '''
    def _validate(self, obj, validator):
        self._coord_keys = ['x', 'y']
        validator.fail_if_key_missing(obj, self._coord_keys)
        if 'z' in obj.columns:
            self._coord_keys.append('z')

    @property
    def coordinates(self):
        '''Returns the coordinate colums of the accessed DataFrame

        Returns
        -------
        coordinates : pandas.DataFrame
            The coordinates `x`, `y` and if 3D `z` of the accessed mesh
        '''
        return self._obj[self._coord_keys]


@pd.api.extensions.register_dataframe_accessor("mesh")
class MeshAccessor(PlainMeshAccessor):

    '''DataFrame accessor to access FEM mesh data (2D and 3D)

    Raises
    ------
    AttributeError
        if at least one of the columns `x`, `y` is missing
    AttributeError
        if the index of the DataFrame is not a two level MultiIndex
        with the names `node_id` and `element_id`

    Notes
    -----
    The MeshAccessor describes how we expect FEM data to look like. It
    consists of nodes identified by `node_id` and elements identified
    by `element_id`. A node playing a role in several elements and an
    element consists of several nodes. So in the DataFrame a `node_id`
    can appear multiple times (for each element, the node is playing a
    role in). Likewise each `element_id` appears multiple times (for
    each node the element consists of).

    The combination `node_id`:`element_id` however, is unique. So the
    table is indexed by a :class:`pandas.MultiIndex` with the level
    names `node_id`, `element_id`.

    See also
    --------
    :class:`PlainMeshAccessor`: accesses meshes without connectivity information
    :func:`pandas.api.extensions.register_dataframe_accessor()`: concept of DataFrame accessors

    Examples
    --------
    For an example see :mod:`meshplot`.
    '''
    def _validate(self, obj, validator):
        super(MeshAccessor, self)._validate(obj, validator)
        if set(obj.index.names) != set(['element_id', 'node_id']):
            raise AttributeError("A mesh needs a pd.MultiIndex with the names `element_id` and `node_id`")



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np
import scipy.stats as stats


class ProbabilityFit:
    def __init__(self, probs, occurrences):
        ''' Fit samples and their estimated occurrences to a lognorm distribution

        Parameters:
        -----------
        probs : array_like
            The estimated cumulated probabilities of the sample values
            (i.e. estimated by func:`pylife.utils.functions.rossow_cumfreqs`)
        occurences : array_like
            the values of the samples
        '''
        if len(probs) != len(occurrences):
            raise ValueError("probs and occurrence arrays must have the same 1D shape.")
        if len(probs) < 2:
            raise ValueError("Need at least two datapoints for probabilities and occurrences.")
        ppf = stats.norm.ppf(probs)
        self._occurrences = np.array(occurrences, dtype=np.float64)
        self._slope, self._intercept, _, _, _ = stats.linregress(np.log10(self._occurrences), ppf)
        self._ppf = ppf


    @property
    def slope(self):
        return self._slope

    @property
    def intercept(self):
        return self._intercept

    @property
    def occurrences(self):
        return self._occurrences

    @property
    def percentiles(self):
        return self._ppf



# Copyright (c) 2020-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Gyöngyvér Kiss"
__maintainer__ = __author__

from h5py.h5t import string_dtype
import numpy as np

from .exceptions import *
from .vmap_dataset import VMAPDataset


class VMAPSection(VMAPDataset):
    def __init__(self, identifier, name, type_id, material, coordinate_system, integration_type, thickness_type):
        super().__init__(identifier)
        self._name = name
        self._type_id = type_id
        self._material = material
        self._coordinate_system = coordinate_system
        self._integration_type = integration_type
        self._thickness_type = thickness_type

    @property
    def attributes(self):
        if self._identifier is None:
            raise (APIUseError("Need to set_identifier() before requesting the attributes."))
        return (self._identifier, self._name, self._type_id, self._material, self._coordinate_system,
                self._integration_type, self._thickness_type)

    @property
    def dtype(self):
        dt_type = np.dtype({"names": ["myIdentifier", "myName", "myType", "myMaterial", "myCoordinateSystem",
                                      "myIntegrationType", "myThicknessType"],
                            "formats": ['<i4', string_dtype(), '<i4', '<i4', '<i4', '<i4', '<i4']})
        return dt_type

    @property
    def dataset_name(self):
        return 'SECTION'



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Gyöngyvér Kiss"
__maintainer__ = __author__

from IPython.display import display

from pylife.materialdata.woehler.controls.radio_button_woehler_curve import RadioButtonWoehlerCurve

class DataFileDisplay(RadioButtonWoehlerCurve):
    def __init__(self, data):
        super().__init__(['Head of the data', 'Details of the data'], 'File display selection')
        self.data = data
        display(self.data.head())

    def selection_changed_handler(self, change):
        self.clear_selection_change_output()

        if change['new'] == change.owner.options[0]:
            display(self.data.head())
        elif change['new'] == change.owner.options[1]:
            display(self.data.describe())
        else:
            raise AttributeError('Unexpected selection')




# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import pandas as pd
import numpy as np
import scipy.stats as stats

from pylife.stress import stresssignal
from pylife.utils.functions import scatteringRange2std


@pd.api.extensions.register_dataframe_accessor('infinite_security')
class InfiniteSecurityAccessor(stresssignal.CyclicStressAccessor):
    """ Compute Security factor for infinite cyclic stress.

    """

    def factors(self, woehler_data, allowed_failure_probability):
        """compute security factor between strength and existing stress.

        Parameters
        ----------
        woehler_data : dict
            strength_inf: double - value for the allowed infinite cyclic strength
            strength_scatter: double - scattering of strength
        allowed_failure_probability : double
            Failure Probability we allow

        Returns
        -------
        double
            quotient of strength and existing stress

        """
        wd = woehler_data
        std_dev = scatteringRange2std(wd.strength_scatter)
        allowed_stress = 10**stats.norm.ppf(allowed_failure_probability,
                                            loc=np.log10(wd.strength_inf),
                                            scale=std_dev)
        factors = allowed_stress / self._obj.sigma_a

        return pd.DataFrame({'security_factor': factors}, index=self._obj.index)



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

from .elementary import Elementary

import numpy as np
import scipy.stats as stats
import pylife.utils.functions as functions
from pylife.utils.probability_data import ProbabilityFit


class Probit(Elementary):
    def _specific_analysis(self, wc):
        self._inf_groups = self._fd.infinite_zone.groupby('load')
        if len(self._inf_groups) < 2:
            warnings.warn(UserWarning("Probit needs at least two runout load levels. Falling back to Elementary."))
            return wc

        wc['1/TS'], wc['SD_50'], wc['ND_50'] = self.__probit_analysis()
        return wc

    def __probit_rossow_estimation(self):
        frac_num = self._inf_groups.fracture.sum().astype(int).to_numpy()
        tot_num = self._inf_groups.fracture.count().to_numpy()

        fprobs = np.empty_like(frac_num, dtype=np.float64)
        c1 = frac_num == 0
        w = np.where(c1)
        fprobs[w] = 1. - 0.5**(1./tot_num[w])

        c2 = frac_num == tot_num
        w = np.where(c2)
        fprobs[w] = 0.5**(1./tot_num[w])

        c3 = np.logical_and(np.logical_not(c1), np.logical_not(c2))
        w = np.where(c3)
        fprobs[w] = (3*frac_num[w] - 1) / (3*tot_num[w] + 1)

        return fprobs, self._inf_groups.load.mean()

    def __probit_analysis(self):
        if self._fd.num_runouts == 0:
            return 1., 0., self._transition_cycles(0.0)

        fprobs, load = self.__probit_rossow_estimation()

        self._probability_fit = ProbabilityFit(fprobs, load)

        TS_inv = functions.std2scatteringRange(1./self._probability_fit.slope)
        SD_50 = 10**(-self._probability_fit.intercept/self._probability_fit.slope)
        ND_50 = self._transition_cycles(SD_50)

        return TS_inv, SD_50, ND_50


    def fitter(self):
        return self._probability_fit



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import numpy as np


class RambergOsgood:
    '''Simple implementation of the Ramberg-Osgood relation

    Parameters
    ----------

    E : float
        Young's Modulus
    K : float
        The strength coefficient
    n : float
        The strain hardening coefficient

    Notes
    -----
    The equation implemented is the one that `Wikipedia
    <https://en.wikipedia.org/wiki/Ramberg%E2%80%93Osgood_relationship#Alternative_Formulations>`__
    refers to as "Alternative Formulation". The parameters `n` and `k` in this
    are formulation are the Hollomon parameters.

    '''

    def __init__(self, E, K, n):
        self._E = E
        self._K = K
        self._n = n

    def strain(self, stress):
        '''Calculate the elastic plastic strain for a given stress

        Parameters
        ----------
        stress : array-like float
            The stress

        Returns
        -------
        strain : array-like float
            The resulting strain

        Raises
        ------
        ValueError if stress is negative
        '''
        stress = np.asarray(stress)
        return stress/self._E + self.plastic_strain(stress)

    def plastic_strain(self, stress):
        '''Calculate the plastic strain for a given stress

        Parameters
        ----------
        stress : array-like float
            The stress

        Returns
        -------
        strain : array-like float
            The resulting plastic strain

        Raises
        ------
        ValueError if stress is negative
        '''
        self._fail_if_negative(stress)
        return np.power(stress/self._K, 1./self._n)

    def delta_strain(self, delta_stress):
        '''Calculate the cyclic Masing strain span for a given stress span

        Parameters
        ----------
        delta_stress : array-like float
            The stress span

        Returns
        -------
        delta_strain : array-like float
            The corresponding stress span

        Raises
        ------
        ValueError if delta_stress is negative

        Notes
        -----
        A Masing like behavior is assumed for the material as described in
        `Kerbgrundkonzept <https://de.wikipedia.org/wiki/Kerbgrundkonzept#Masing-Verhalten_und_Werkstoffged%C3%A4chtnis>`__.
        '''
        self._fail_if_negative(delta_stress)
        return delta_stress/self._E + 2.*np.power(delta_stress/(2.*self._K), 1./self._n)

    def lower_hysteresis(self, stress, max_stress):
        '''Calculate the lower (relaxation to compression) hysteresis starting from a given maximum stress

        Parameters
        ----------
        stress : array-like float
            The stress (must be below the maximum stress)
        max_stress : float
            The maximum stress of the hysteresis look

        Returns
        -------
        lower_hysteresis : array-like float
            The lower hysteresis branch from `max_stress` all the way to `stress`

        Raises
        ------
        ValueError if stress > max_stress
        '''
        stress = np.asarray(stress)
        if (stress > max_stress).any():
            raise ValueError("Value for 'stress' must not be higher than 'max_stress'.")
        return self.strain(max_stress) - self.delta_strain(max_stress-stress)

    def _fail_if_negative(self, val):
        if (np.asarray(val) < 0).any():
            raise ValueError("Stress value in Ramberg-Osgood equation must not be negative.")



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Utility Functions
=================

A collection of functions frequently used in lifetime estimation
business.
'''

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import numpy as np


def scatteringRange2std(T_inv):
    ''' Converts a inverted scattering range (1/T) into standard deviation

    Parameters
    ----------
    T_inv : float
        inverted scattering range

    Returns
    -------
    std : float
        standard deviation corresponding to 1/T assuming a normal distribution

    Notes
    -----
    Actually `1/(2*norm.ppf(0.9))*np.log10(T_inv)`

    Inverse of `std2scatteringRange()`
    '''
    return 0.39015207303618954*np.log10(T_inv)


def std2scatteringRange(std):
    ''' Converts a standard deviation into inverted scattering range (1/T)

    Parameters
    ----------
    std : float
        standard deviation

    Returns
    -------
    T_inv : float
        inverted scattering range corresponding to `std` assuming a normal distribution

    Notes
    -----
    Actually `10**(2*norm.ppf(0.9)*std`

    Inverse of `scatteringRange2std()`
    '''
    return 10**(2.5631031310892007*(std))


def rossow_cumfreqs(N):
    """Cumulative frequency estimator according to Rossow

    Parameters
    ----------
    N : int
        The sample size of the statistical population

    Returns
    -------
    cumfreqs : numpy.ndarray
        The estimated cumulated frequencies of the N samples

    Notes
    -----
    The returned value is the probability that the next taken sample
    is below the value of the i-th sample of n sorted samples.

    Examples
    --------
    >>> rossow_cumfreqs(1)
    array([0.5])

    If we have one sample, the probability that the next sample will
    be below it is 0.5.

    >>> rossow_cumfreqs(3)
    array([0.2, 0.5, 0.8])

    If we have three sorted samples, the probability that the next
    sample will be
    * below the first is 0.2
    * below the second is 0.5
    * below the third is 0.8


    References
    ----------
    'Statistics of Metal Fatigue in Engineering' page 16

    https://books.google.de/books?isbn=3752857722

    """
    i = np.arange(1, N+1)
    return (3.*i-1.)/(3.*N+1)



import setuptools, sys, os
import distutils

use_cython = False

if sys.platform == 'win32':
    try:
        msvc = setuptools.msvc.EnvironmentInfo('win32')
        use_cython = True
    except distutils.errors.DistutilsPlatformError:
        pass
else:
    use_cython = True

try:
    from Cython.Build import cythonize
    from Cython.Distutils import build_ext
except ImportError:
    def cythonize (*args, ** kwargs ):
        from Cython.Build import cythonize
        return cythonize(*args, ** kwargs)
    use_cython = False

import versioneer
version=versioneer.get_version()
cmdclass=versioneer.get_cmdclass()

if use_cython:
    cmdclass.update(build_ext=build_ext)

with open("README.md", "r", encoding='utf-8') as fh:
    long_description = fh.read()

cython_modules = [
    setuptools.Extension('pylife.stress.rainflow', sources=['pylife/stress/rainflow.py'])
]

if os.getenv('NO_INSTALL_REQUIREMENTS') != 'true':
    with open("requirements.txt", 'r') as fh:
        requirements = fh.readlines()
else:
    requirements = []

setuptools.setup(
    name="pylife",
    version=version,
    cmdclass=cmdclass,
    author="pyLife developer team @ Bosch Research",
    author_email="Johannes.Mueller4@de.bosch.com",
    description="General Fatigue library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/boschresearch/pylife",
    packages=setuptools.find_packages(exclude=['demos', 'tests.*', 'tests']),
    ext_modules = cythonize(cython_modules, compiler_directives={'linetrace': True}) if use_cython else cython_modules,
    include_package_data=True,
	classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: Apache Software License",
            "Operating System :: OS Independent",
    ],
    setup_requires = [
        'setuptools',
        'cython'
    ],
    install_requires = [ requirements ]
)



# Copyright (c) 2019-2021 - for information on the respective copyright owner
# see the NOTICE file and/or the repository
# https://github.com/boschresearch/pylife
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Johannes Mueller"
__maintainer__ = __author__

import pandas as pd
import numpy as np

from pylife import signal


@pd.api.extensions.register_dataframe_accessor("voigt")
class StressTensorVoigtAccessor(signal.PylifeSignal):
    '''DataFrame accessor class for Voigt noted stress tensors

    Raises
    ------
    AttributeError
        if at least one of the needed columns is missing.

    Notes
    -----
    Base class to access :class:`pandas.DataFrame` objects containing
    Voigt noted stress tensors. The stress tensor components are assumed
    to be in the columns `S11`, `S22`, `S33`, `S12`, `S13`, `S23`.

    See also
    --------
    :func:`pandas.api.extensions.register_dataframe_accessor()`

    Examples
    --------
    For an example see :class:`equistress.StressTensorEquistressAccessor`.
    '''
    def _validate(self, obj, validator):
        validator.fail_if_key_missing(obj, ['S11', 'S22', 'S33', 'S12', 'S13', 'S23'])


@pd.api.extensions.register_dataframe_accessor("cyclic_stress")
class CyclicStressAccessor(signal.PylifeSignal):
    '''DataFrame accessor class for cyclic stress data

    Raises
    ------
    AttributeError
        if there is no stress amplitide `sigma_a` key in the data

    Notes
    -----
    Base class to access :class:`pandas.DataFrame` objects containing
    cyclic stress data consisting of at least a stress amplitude using
    the key `sigma_a`. Optionally an value for the stress ratio `R` or
    for meanstress `sigma_m` can be supplied. If neither `R` nor
    `sigma_a` are supplied a mean stress of zero i.e. `R=-1` is assumed.

    Todo
    ----
    Handle also input data with lower and upper stress.
    '''
    def _validate(self, obj, validator):

        if 'sigma_a' in obj.columns and 'R' in obj.columns:
            obj['sigma_m'] = obj['sigma_a']*(1.+obj.R)/(1.-obj.R)
            obj.loc[obj['R'] == -np.inf, 'sigma_m'] = -obj.sigma_a
            return True
        if 'sigma_a' in obj.columns and 'sigma_m' in obj.columns:
            return True
        if 'sigma_a' in obj.columns:
            obj['sigma_m'] = np.zeros_like(obj['sigma_a'].to_numpy())
        else:
            validator.fail_if_key_missing(obj, ['sigma_m', 'sigma_a'])
        return True

    def constant_R(self, R):
        ''' Sets `sigma_m` in a way that `R` is a constant value.

        Parameters
        ----------
        R : float
            The value for `R`.

        Returns
        -------
        the accessed DataFrame : :class:`pandas.DataFrame`
        '''
        self._obj['sigma_m'] = self._obj['sigma_a']*(1.+R)/(1.-R)
        return self._obj
